"""
The wntr.graphics.network module includes methods plot the
water network model.
"""
import logging
import networkx as nx
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import animation

try:
    import plotly
except:
    plotly = None
try:
    import folium
except:
    folium = None
    
from wntr.graphics.color import custom_colormap

logger = logging.getLogger(__name__)

def _format_node_attribute(node_attribute, wn):
    
    if isinstance(node_attribute, str):
        node_attribute = wn.query_node_attribute(node_attribute)
    if isinstance(node_attribute, list):
        node_attribute = dict(zip(node_attribute,[1]*len(node_attribute)))
    if isinstance(node_attribute, pd.Series):
        node_attribute = dict(node_attribute)
    
    return node_attribute

def _format_link_attribute(link_attribute, wn):
    
    if isinstance(link_attribute, str):
        link_attribute = wn.query_link_attribute(link_attribute)
    if isinstance(link_attribute, list):
        link_attribute = dict(zip(link_attribute,[1]*len(link_attribute)))
    if isinstance(link_attribute, pd.Series):
        link_attribute = dict(link_attribute)
            
    return link_attribute
        
def plot_network(wn, node_attribute=None, link_attribute=None, title=None,
               node_size=20, node_range=[None,None], node_alpha=1, node_cmap=None, node_labels=False,
               link_width=1, link_range=[None,None], link_alpha=1, link_cmap=None, link_labels=False,
               valve_layer=None, add_colorbar=True, node_colorbar_label='Node', link_colorbar_label='Link', 
               directed=False, ax=None, filename=None):
    """
    Plot network graphic

    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object

    node_attribute : None, str, list, pd.Series, or dict, optional

        - If node_attribute is a string, then a node attribute dictionary is
          created using node_attribute = wn.query_node_attribute(str)
        - If node_attribute is a list, then each node in the list is given a 
          value of 1.
        - If node_attribute is a pd.Series, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float. 
        - If node_attribute is a dict, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float

    link_attribute : None, str, list, pd.Series, or dict, optional

        - If link_attribute is a string, then a link attribute dictionary is
          created using edge_attribute = wn.query_link_attribute(str)
        - If link_attribute is a list, then each link in the list is given a 
          value of 1.
        - If link_attribute is a pd.Series, then it should be in the format
          {linkid: x} where linkid is a string and x is a float. 
        - If link_attribute is a dict, then it should be in the format
          {linkid: x} where linkid is a string and x is a float.

    title : str, optional
        Plot title 

    node_size : int, optional
        Node size 

    node_range : list, optional
        Node range ([None,None] indicates autoscale)
        
    node_alpha : int, optional
        Node transparency
        
    node_cmap : matplotlib.pyplot.cm colormap or list of named colors, optional
        Node colormap 
        
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name. 
        
    link_width : int, optional
        Link width

    link_range : list, optional
        Link range ([None,None] indicates autoscale)

    link_alpha : int, optional
        Link transparency
    
    link_cmap : matplotlib.pyplot.cm colormap or list of named colors, optional
        Link colormap
        
    link_labels: bool, optional
        If True, the graph will include each link labelled with its name. 
        
    add_colorbar : bool, optional
        Add colorbar

    node_colorbar_label: str, optional
        Node colorbar label
        
    link_colorbar_label: str, optional
        Link colorbar label
        
    directed : bool, optional
        If True, plot the directed graph
    
    ax : matplotlib axes object, optional
        Axes for plotting (None indicates that a new figure with a single 
        axes will be used)
        
    Returns
    -------
    nodes, edges : matplotlib objects for network nodes and edges

    Notes
    -----
    For more network draw options, see nx.draw_networkx
    """

    if ax is None: # create a new figure
        plt.figure(facecolor='w', edgecolor='k')
        ax = plt.gca()
        
    # Graph
    G = wn.get_graph()
    if not directed:
        G = G.to_undirected()

    # Position
    pos = nx.get_node_attributes(G,'pos')
    if len(pos) == 0:
        pos = None

    # Define node properties
    add_node_colorbar = add_colorbar
    if node_attribute is not None:
        
        if isinstance(node_attribute, list):
            if node_cmap is None:
                node_cmap = ['red', 'red']
            add_node_colorbar = False
        
        if node_cmap is None:
            node_cmap = plt.cm.Spectral_r
        elif isinstance(node_cmap, list):
            if len(node_cmap) == 1:
                node_cmap = node_cmap*2
            node_cmap = custom_colormap(len(node_cmap), node_cmap)  
         
        node_attribute = _format_node_attribute(node_attribute, wn)
        nodelist,nodecolor = zip(*node_attribute.items())

    else:
        nodelist = None
        nodecolor = 'k'
    
    add_link_colorbar = add_colorbar
    if link_attribute is not None:
        
        if isinstance(link_attribute, list):
            if link_cmap is None:
                link_cmap = ['red', 'red']
            add_link_colorbar = False

        if link_cmap is None:
            link_cmap = plt.cm.Spectral_r
        elif isinstance(link_cmap, list):
            if len(link_cmap) == 1:
                link_cmap = link_cmap*2
            link_cmap = custom_colormap(len(link_cmap), link_cmap)  
            
        link_attribute = _format_link_attribute(link_attribute, wn)
        
        # Replace link_attribute dictionary defined as
        # {link_name: attr} with {(start_node, end_node, link_name): attr}
        attr = {}
        for link_name, value in link_attribute.items():
            link = wn.get_link(link_name)
            attr[(link.start_node_name, link.end_node_name, link_name)] = value
        link_attribute = attr
        
        linklist,linkcolor = zip(*link_attribute.items())
    else:
        linklist = None
        linkcolor = 'k'
    
    if title is not None:
        ax.set_title(title)
        
    edge_background = nx.draw_networkx_edges(G, pos, edge_color='grey', 
                                             width=0.5, ax=ax)
    
    nodes = nx.draw_networkx_nodes(G, pos, 
            nodelist=nodelist, node_color=nodecolor, node_size=node_size, 
            alpha=node_alpha, cmap=node_cmap, vmin=node_range[0], vmax = node_range[1], 
            linewidths=0, ax=ax)
    edges = nx.draw_networkx_edges(G, pos, edgelist=linklist, 
            edge_color=linkcolor, width=link_width, alpha=link_alpha, edge_cmap=link_cmap, 
            edge_vmin=link_range[0], edge_vmax=link_range[1], ax=ax)
    if node_labels:
        labels = dict(zip(wn.node_name_list, wn.node_name_list))
        nx.draw_networkx_labels(G, pos, labels, font_size=7, ax=ax)
    if link_labels:
        labels = {}
        for link_name in wn.link_name_list:
            link = wn.get_link(link_name)
            labels[(link.start_node_name, link.end_node_name)] = link_name
        nx.draw_networkx_edge_labels(G, pos, labels, font_size=7, ax=ax)
    if add_node_colorbar and node_attribute:
        clb = plt.colorbar(nodes, shrink=0.5, pad=0, ax=ax)
        clb.ax.set_title(node_colorbar_label, fontsize=10)
    if add_link_colorbar and link_attribute:
        clb = plt.colorbar(edges, shrink=0.5, pad=0.05, ax=ax)
        clb.ax.set_title(link_colorbar_label, fontsize=10)
    ax.axis('off')
    
    if valve_layer is not None:
        for valve_name, (pipe_name, node_name) in valve_layer.iterrows():
            pipe = wn.get_link(pipe_name)
            if node_name == pipe.start_node_name:
                start_node = pipe.start_node
                end_node = pipe.end_node
            elif node_name == pipe.end_node_name:
                start_node = pipe.end_node
                end_node = pipe.start_node
            else:
                print("Not valid")
                continue
            x0 = start_node.coordinates[0]
            dx = end_node.coordinates[0] - x0
            y0 = start_node.coordinates[1]
            dy = end_node.coordinates[1] - y0
            valve_coordinates = (x0 + dx * 0.1,
                                     y0 + dy * 0.1)
            ax.scatter(valve_coordinates[0], valve_coordinates[1], 15, 'r', 'v')   
     
    if filename:
        plt.savefig(filename)
        
    return nodes, edges

def plot_interactive_network(wn, node_attribute=None, node_attribute_name = 'Value', title=None,
               node_size=8, node_range=[None,None], node_cmap='Jet', node_labels=True,
               link_width=1, add_colorbar=True, reverse_colormap=False,
               figsize=[700, 450], round_ndigits=2, add_to_node_popup=None, 
               filename='plotly_network.html', auto_open=True):
    """
    Create an interactive scalable network graphic using plotly.  

    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object

    node_attribute : None, str, list, pd.Series, or dict, optional

        - If node_attribute is a string, then a node attribute dictionary is
          created using node_attribute = wn.query_node_attribute(str)
        - If node_attribute is a list, then each node in the list is given a 
          value of 1.
        - If node_attribute is a pd.Series, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float.
          The time index is not used in the plot.
        - If node_attribute is a dict, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float

    node_attribute_name : str, optional 
        The node attribute name, which is used in the node popup and node legend

    title : str, optional
        Plot title

    node_size : int, optional
        Node size

    node_range : list, optional
        Node range ([None,None] indicates autoscale)

    node_cmap : palette name string, optional
        Node colormap, options include Greys, YlGnBu, Greens, YlOrRd, Bluered, 
        RdBu, Reds, Blues, Picnic, Rainbow, Portland, Jet, Hot, Blackbody, 
        Earth, Electric, Viridis
    
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name and 
        attribute value.
        
    link_width : int, optional
        Link width
    
    add_colorbar : bool, optional
        Add colorbar
        
    reverse_colormap : bool, optional
        Reverse colormap
        
    figsize: list, optional
        Figure size in pixels

    round_ndigits : int, optional
        Number of digits to round node values used in the label
    
    add_to_node_popup : None or pd.DataFrame, optional
        To add additional information to the node popup, use a DataFrame with 
        node name as index and attributes as values.  Column names will be added
        to the popup along with each value for a given node.
        
    filename : string, optional
        HTML file name
    
    auto_open : bool, optional
        Open the HTML file after creation
    """
    if plotly is None:
        raise ImportError('plotly is required')
        
    # Graph
    G = wn.get_graph()
    
    # Node attribute
    if node_attribute is not None:
        if isinstance(node_attribute, list):
            node_cmap = 'Reds'
            add_colorbar = False
        node_attribute = _format_node_attribute(node_attribute, wn)
    else:
        add_colorbar = False
        
    # Create edge trace
    edge_trace = plotly.graph_objs.Scatter(
        x=[], 
        y=[], 
        text=[],
        hoverinfo='text',
        mode='lines',
        line=dict(
            #colorscale=link_cmap,
            #reversescale=reverse_colormap,
            color='#888', #[], 
            width=link_width))
    for edge in G.edges():
        x0, y0 = G.nodes[edge[0]]['pos']
        x1, y1 = G.nodes[edge[1]]['pos']
        edge_trace['x'] += tuple([x0, x1, None])
        edge_trace['y'] += tuple([y0, y1, None])
#        try:
#            # Add link attributes
#            link_name = G[edge[0]][edge[1]].keys()[0]
#            edge_trace['line']['color'] += tuple([pipe_attr[link_name]])
#            edge_info = 'Edge ' + str(link_name)
#            edge_trace['text'] += tuple([edge_info])
#        except:
#            pass
#    edge_trace['colorbar']['title'] = 'Link colorbar title'
    
    # Create node trace      
    node_trace = plotly.graph_objs.Scatter(
        x=[], 
        y=[], 
        text=[],
        hoverinfo='text',
        mode='markers', 
        marker=dict(
            showscale=add_colorbar,
            colorscale=node_cmap, 
            cmin=node_range[0], # TODO: Not sure this works
            cmax=node_range[1], # TODO: Not sure this works
            reversescale=reverse_colormap,
            color=[], 
            size=node_size,         
            #opacity=0.75,
            colorbar=dict(
                thickness=15,
                xanchor='left',
                titleside='right'),
            line=dict(width=1)))
    for node in G.nodes():
        x, y = G.nodes[node]['pos']
        node_trace['x'] += tuple([x])
        node_trace['y'] += tuple([y])
        try:
            # Add node attributes
            node_trace['marker']['color'] += tuple([node_attribute[node]])
            #node_trace['marker']['size'].append(node_size)

            # Add node labels
            if node_labels:
                node_info = wn.get_node(node).node_type + ': ' + str(node) + '<br>'+ \
                            node_attribute_name + ': ' + str(round(node_attribute[node],round_ndigits))
                if add_to_node_popup is not None:
                    if node in add_to_node_popup.index:
                        for key, val in add_to_node_popup.loc[node].iteritems():
                            node_info = node_info + '<br>' + \
                                key + ': ' + '{:.{prec}f}'.format(val, prec=round_ndigits)
                            
                node_trace['text'] += tuple([node_info])
        except:
            node_trace['marker']['color'] += tuple(['#888'])
            if node_labels:
                node_info = wn.get_node(node).node_type + ': ' + str(node)
                
                node_trace['text'] += tuple([node_info])
            #node_trace['marker']['size'] += tuple([5])
    #node_trace['marker']['colorbar']['title'] = 'Node colorbar title'
    
    # Create figure
    data = [edge_trace, node_trace]
    layout = plotly.graph_objs.Layout(
                    title=title,
                    titlefont=dict(size=16),
                    showlegend=False, 
                    width=figsize[0],
                    height=figsize[1],
                    hovermode='closest',
                    margin=dict(b=20,l=5,r=5,t=40),
                    xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                    yaxis=dict(showgrid=False, zeroline=False, showticklabels=False))
    
    fig = plotly.graph_objs.Figure(data=data,layout=layout)
    if filename:
        plotly.offline.plot(fig, filename=filename, auto_open=auto_open)  
    else:
        plotly.offline.plot(fig, auto_open=auto_open)  

def plot_leaflet_network(wn, node_attribute=None, link_attribute=None, 
               node_attribute_name = 'Value', 
               link_attribute_name = 'Value',
               node_size=2, node_range=[None,None], 
               node_cmap=['cornflowerblue', 'forestgreen', 'gold', 'firebrick'], 
               node_cmap_bins = 'cut', node_labels=True,
               link_width=2, link_range=[None,None], 
               link_cmap=['cornflowerblue', 'forestgreen', 'gold', 'firebrick'], 
               link_cmap_bins='cut', link_labels=True,
               add_legend=False, round_ndigits=2, zoom_start=13, 
               add_to_node_popup=None, add_to_link_popup=None,
               filename='leaflet_network.html'):
    """
    Create an interactive scalable network graphic on a Leaflet map using folium.  

    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object

    node_attribute : None, str, list, pd.Series, or dict, optional

        - If node_attribute is a string, then a node attribute dictionary is
          created using node_attribute = wn.query_node_attribute(str)
        - If node_attribute is a list, then each node in the list is given a 
          value of 1.
        - If node_attribute is a pd.Series, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float. 
        - If node_attribute is a dict, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float

    link_attribute : None, str, list, pd.Series, or dict, optional

        - If link_attribute is a string, then a link attribute dictionary is
          created using edge_attribute = wn.query_link_attribute(str)
        - If link_attribute is a list, then each link in the list is given a 
          value of 1.
        - If link_attribute is a pd.Series, then it should be in the format
          {linkid: x} where linkid is a string and x is a float. 
        - If link_attribute is a dict, then it should be in the format
          {linkid: x} where linkid is a string and x is a float.

    node_attribute_name : str, optional 
        The node attribute name, which is used in the node popup and node legend
        
    link_attribute_name : str, optional 
        The link attribute name, which is used in the link popup and link legend
        
    node_size : int, optional
        Node size 

    node_range : list, optional
        Node range ([None,None] indicates autoscale)

    node_cmap : list of color names, optional
        Node colors 
    
    node_cmap_bins: string, optional
        Node color bins, 'cut' or 'qcut'
    
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name. 
        
    link_width : int, optional
        Link width

    link_range : list, optional
        Link range ([None,None] indicates autoscale)

    link_cmap : list of color names, optional
        Link colors
    
    link_cmap_bins: string, optional
        Link color bins, 'cut' or 'qcut'
        
    link_labels: bool, optional
        If True, the graph will include each link labelled with its name. 
    
    add_legend: bool, optional
         Add a legend to the map
    
    round_ndigits : int, optional
        Rounds digits in the popup
        
    zoom_start : int, optional
        Zoom start used to set initial scale of the map
    
    add_to_node_popup : None or pd.DataFrame, optional
        To add additional information to the node popup, use a DataFrame with 
        node name as index and attributes as values.  Column names will be added
        to the popup along with each value for a given node.
        
    add_to_link_popup : None or pd.DataFrame, optional
        To add additional information to the link popup, use a DataFrame with 
        link name as index and attributes as values.  Column names will be added
        to the popup along with each value for a given link.
        
    filename : str, optional
        Filename used to save the map
    """
    
    if folium is None:
        raise ImportError('folium is required')
    
    if node_attribute is not None:
        if isinstance(node_attribute, list):
            node_cmap=['red']
        node_attribute = _format_node_attribute(node_attribute, wn)
        node_attribute = pd.Series(node_attribute)
        if node_range[0] is not None:
            node_attribute[node_attribute < node_range[0]] = node_range[0]
        if node_range[1] is not None:
            node_attribute[node_attribute > node_range[1]] = node_range[1]
        if node_cmap_bins == 'cut':
            node_colors, node_bins = pd.cut(node_attribute, len(node_cmap), 
                                                labels=node_cmap, retbins =True)
        elif node_cmap_bins == 'qcut':
            node_colors, node_bins = pd.qcut(node_attribute, len(node_cmap), 
                                                 labels=node_cmap, retbins =True)
        
    if link_attribute is not None:
        if isinstance(link_attribute, list):
            link_cmap=['red']
        link_attribute = _format_link_attribute(link_attribute, wn)
        link_attribute = pd.Series(link_attribute)
        if link_range[0] is not None:
            link_attribute[link_attribute < link_range[0]] = link_range[0]
        if link_range[1] is not None:
            link_attribute[link_attribute > link_range[1]] = link_range[1]
        if link_cmap_bins == 'cut':
            link_colors, link_bins  = pd.cut(link_attribute, len(link_cmap), 
                                             labels=link_cmap, retbins =True)
        elif link_cmap_bins == 'qcut':
            link_colors, link_bins  = pd.qcut(link_attribute, len(link_cmap), 
                                              labels=link_cmap, retbins =True)
        
    G = wn.get_graph()
    pos = nx.get_node_attributes(G,'pos')
    center = pd.DataFrame(pos).mean(axis=1)
    
    m = folium.Map(location=[center.iloc[1], center.iloc[0]], zoom_start=zoom_start, 
                   tiles='cartodbpositron')
    #folium.TileLayer('cartodbpositron').add_to(m)
    
    # Node popup
    node_popup = {k: '' for k in wn.node_name_list}
    if node_labels:
        for name, node in wn.nodes():
            node_popup[name] = node.node_type + ': ' + name
            if node_attribute is not None:
                if name in node_attribute.index:
                    node_popup[name] = node_popup[name] + '<br>' + \
                        node_attribute_name + ': ' + '{:.{prec}f}'.format(node_attribute[name], prec=round_ndigits)
            if add_to_node_popup is not None:
                if name in add_to_node_popup.index:
                    for key, val in add_to_node_popup.loc[name].iteritems():
                        node_popup[name] = node_popup[name] + '<br>' + \
                            key + ': ' + '{:.{prec}f}'.format(val, prec=round_ndigits)
                 
    # Link popup
    link_popup = {k: '' for k in wn.link_name_list}
    if link_labels:
        for name, link in wn.links():
            link_popup[name] = link.link_type + ': ' + name
            if link_attribute is not None:
                if name in link_attribute.index:
                    link_popup[name] = link_popup[name] + '<br>' + \
                        link_attribute_name + ': ' + '{:.{prec}f}'.format(link_attribute[name], prec=round_ndigits)
            if add_to_link_popup is not None:
                if name in add_to_link_popup.index:
                    for key, val in add_to_link_popup.loc[name].iteritems():
                        link_popup[name] = link_popup[name] + '<br>' + \
                            key + ': ' + '{:.{prec}f}'.format(val, prec=round_ndigits)
                            
    if node_size > 0:
        for name, node in wn.nodes():
            loc = (node.coordinates[1], node.coordinates[0])
            radius = node_size
            color = 'black'
            if node_labels:
                popup = node_popup[name]
            else:
                popup = None
                    
            if node_attribute is not None:
                if name in node_attribute.index:
                    color = node_colors[name]
                else:
                    radius = 0.1
            
            folium.CircleMarker(loc, popup=popup, color=color, fill=True, 
                                fill_color=color, radius=radius, fill_opacity=0.7, opacity=0.7).add_to(m)
            
    if link_width > 0:
        for name, link in wn.links():            
            start_loc = (link.start_node.coordinates[1], link.start_node.coordinates[0])
            end_loc = (link.end_node.coordinates[1], link.end_node.coordinates[0])
            weight = link_width
            color='black'
            if link_labels:
                popup = link_popup[name]
            else:
                popup = None
            
            if link_attribute is not None:
                if name in link_attribute.index:
                    color = link_colors[name]
                else:
                    weight = 1.5
            
            folium.PolyLine([start_loc, end_loc], popup=popup, color=color, 
                            weight=weight, opacity=0.7).add_to(m)
    
    if (add_legend) & ((len(node_cmap) >= 1) or (len(link_cmap) >= 1)):
        if node_attribute is not None:  #Produce node legend
            height = 50+len(node_cmap)*20 + (int(len(node_attribute_name)/20) + 1)*20
            node_legend_html = """<div style="position: fixed; 
        bottom: 50px; left: 50px; width: 150px; height: """+str(height)+"""px; 
        background-color:white;z-index:9999; font-size:14px; "><br>
            <b><P ALIGN=CENTER>""" + "Node Legend: " + node_attribute_name + """</b> </P>"""
            for color, val in zip(node_cmap, node_bins[0:-1]):
                val = '{:.{prec}f}'.format(val, prec=round_ndigits)
                node_legend_html += """
                &emsp;<i class="fa fa-circle fa-1x" 
                style="color:"""+ color +""" "></i> >= """+ val +""" <br>"""
            node_legend_html += """</div>"""
            m.get_root().html.add_child(folium.Element(node_legend_html))
			
        if link_attribute is not None:   #Produce link legend
            height = 50+len(link_cmap)*20 + (int(len(link_attribute_name)/20) + 1)*20
            link_legend_html = """<div style="position: fixed; 
			bottom: 50px; left: 250px; width: 150px; height: """+str(height)+"""px; 
			background-color:white;z-index:9999; font-size:14px; "><br>
            <b><P ALIGN=CENTER>""" + "Link Legend: " + link_attribute_name + """</b> </P>"""
            for color, val in zip(link_cmap, link_bins[0:-1]):
                val = '{:.{prec}f}'.format(val, prec=round_ndigits)
                link_legend_html += """
               &emsp;<i class="fa fa-minus fa-1x" 
                style="color:"""+ color +""" "></i> >= """+ val +""" <br>"""
            link_legend_html += """</div>"""
            m.get_root().html.add_child(folium.Element(link_legend_html))
    
    #plugins.Search(points, search_zoom=20, ).add_to(m)
    #if add_longlat_popup:
    #    m.add_child(folium.LatLngPopup())
    
    folium.LayerControl().add_to(m)
    
    m.save(filename)
 
def network_animation(wn, node_attribute=None, link_attribute=None, title=None,
               node_size=20, node_range=[None,None], node_alpha=1, node_cmap=None, node_labels=False,
               link_width=1, link_range=[None,None], link_alpha=1, link_cmap=None, link_labels=False,
               add_colorbar=True, directed=False, ax=None, repeat=True):
    """
    Create a network animation

    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object

    node_attribute : pd.DataFrame, optional
        Node attributes stored in a pandas DataFrames, where the index is 
        time and columns are the node name 

    link_attribute : pd.DataFrame, optional
        Link attributes stored in a pandas DataFrames, where the index is 
        time and columns are the link name 

    title : str, optional
        Plot title 

    node_size : int, optional
        Node size 

    node_range : list, optional
        Node range ([None,None] indicates autoscale)
        
    node_alpha : int, optional
        Node transparency
        
    node_cmap : matplotlib.pyplot.cm colormap or list of named colors, optional
        Node colormap 
        
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name. 
        
    link_width : int, optional
        Link width

    link_range : list, optional
        Link range ([None,None] indicates autoscale)

    link_alpha : int, optional
        Link transparency
    
    link_cmap : matplotlib.pyplot.cm colormap or list of named colors, optional
        Link colormap
        
    link_labels: bool, optional
        If True, the graph will include each link labelled with its name. 
        
    add_colorbar : bool, optional
        Add colorbar

    directed : bool, optional
        If True, plot the directed graph
    
    repeat : bool, optional
        If True, the animation will repeat
        
    Returns
    -------
    matplotlib animation
    """
    
    if node_attribute is not None:
        node_index = node_attribute.index
        initial_node_values = node_attribute.iloc[0, :]
        if node_range[0] is None:
            node_range[0] = node_attribute.min().min()
        if node_range[1] is None:
            node_range[1] = node_attribute.max().max()
    else:
        node_index = None
        initial_node_values = None
        
    if link_attribute is not None:
        link_index = link_attribute.index
        initial_link_values = link_attribute.iloc[0, :]
        if link_range[0] is None:
            link_range[0] = link_attribute.min().min()
        if link_range[1] is None:
            link_range[1] = link_attribute.max().max()
    else:
        link_index = None
        initial_link_values = None
    
    if (node_index is not None) & (link_index is not None):
        if len(node_index.symmetric_difference(link_index)) > 0:
            print('Node attribute index does not equal link attribute index')
            return
        index = node_index
    elif node_index is not None:
        index = node_index
    elif link_index is not None:
        index = link_index
    else:
        print('Node attributes or link attributes must be included')
        return
    
    if ax is None: # create a new figure
        fig = plt.figure(facecolor='w', edgecolor='k')
        ax = plt.gca()
            
    if title is not None:
        title_name = title + ', ', str(index[0])
    else:
        title_name = '0'
    
    nodes, edges = plot_network(wn, node_attribute=initial_node_values, link_attribute=initial_link_values, title=title_name,
               node_size=node_size, node_range=node_range, node_alpha=node_alpha, node_cmap=node_cmap, node_labels=node_labels,
               link_width=link_width, link_range=link_range, link_alpha=link_alpha, link_cmap=link_cmap, link_labels=link_labels,
               add_colorbar=add_colorbar, directed=directed, ax=ax)
        
    def update(n):
        if node_attribute is not None:
            node_values = node_attribute.iloc[n, :]
        else:
            node_values = None
        
        if link_attribute is not None:
            link_values = link_attribute.iloc[n, :]
        else:
            link_values = None
            
        if title is not None:
            title_name = title + ', ' + str(index[n])
        else:
            title_name = str(n)
        
        fig.clf()  
        ax = plt.gca()
        
        nodes, edges = plot_network(wn, node_attribute=node_values, link_attribute=link_values, title=title_name,
               node_size=node_size, node_range=node_range, node_alpha=node_alpha, node_cmap=node_cmap, node_labels=node_labels,
               link_width=link_width, link_range=link_range, link_alpha=link_alpha, link_cmap=link_cmap, link_labels=link_labels,
               add_colorbar=add_colorbar, directed=directed, ax=ax)
        
        return nodes, edges
    
    anim = animation.FuncAnimation(fig, update, interval=50, frames=len(index), blit=False, repeat=repeat)
    
    return anim


"""
The wntr.epanet.util module contains unit conversion utilities based on EPANET units.

.. rubric:: Contents

.. autosummary::

    FlowUnits
	MassUnits
	QualParam
	HydParam
	to_si
	from_si
	StatisticsType
	QualType
	SourceType
	PressureUnits
	FormulaType
	ControlType
	LinkTankStatus
	MixType
	ResultType
	EN

"""
import numpy as np
import pandas as pd
import enum
import logging

logger = logging.getLogger(__name__)

__all__ = ["FlowUnits", "MassUnits", "QualParam", "HydParam", "to_si", "from_si",
           "StatisticsType", "QualType", "SourceType", "PressureUnits", "FormulaType",
           "ControlType", "LinkTankStatus",
           "MixType", "ResultType", "EN"]


class FlowUnits(enum.Enum):
    u"""Epanet Units Enum class.

    EPANET has defined unit codes that are used in its INP input files.
    This enumerated type class provides the appropriate values, rather than
    setting up a large number of constants. Additionally, each Enum value has
    a property that identifies it as either `traditional` or `metric` flow unit.
    EPANET *does not* use fully SI units - these are provided for WNTR compatibility.

    .. rubric:: Enum Members

    ==============  ====================================  ========================
    :attr:`~CFS`    :math:`ft^3\,/\,s`                    :attr:`is_traditional`
    :attr:`~GPM`    :math:`gal\,/\,min`                   :attr:`is_traditional`
    :attr:`~MGD`    :math:`10^6\,gal\,/\,day`             :attr:`is_traditional`
    :attr:`~IMGD`   :math:`10^6\,Imp.\,gal\,/\,day`       :attr:`is_traditional`
    :attr:`~AFD`    :math:`acre\cdot\,ft\,/\,day`         :attr:`is_traditional`
    :attr:`~LPS`    :math:`L\,/\,s`                       :attr:`is_metric`
    :attr:`~LPM`    :math:`L\,/\,min`                     :attr:`is_metric`
    :attr:`~MLD`    :math:`ML\,/\,day`                    :attr:`is_metric`
    :attr:`~CMH`    :math:`m^3\,\,hr`                     :attr:`is_metric`
    :attr:`~CMD`    :math:`m^3\,/\,day`                   :attr:`is_metric`
    :attr:`~SI`     :math:`m^3\,/\,s`
    ==============  ====================================  ========================

    .. rubric:: Enum Member Attributes

    .. autosummary::
        factor
        is_traditional
        is_metric


    Examples
    --------
    >>> from wntr.epanet import FlowUnits
    >>> FlowUnits.GPM
    <FlowUnits.GPM: (1, 6.30901964e-05)>

    Units can be converted to the EPANET integer values by casting as an ``int`` and can be
    converted to a string by accessing the ``name`` property. The factor to convert to SI units
    is accessed using the ``factor`` property.

    >>> FlowUnits.LPS.name
    'LPS'
    >>> int(FlowUnits.LPS)
    5

    The reverse is also true, where an ``int`` from an EPANET run or the string from and input
    file can be used to get a ``FlowUnits`` object.

    >>> FlowUnits(4)
    <FlowUnits.AFD: (4, 0.014276410185185185)>
    >>> FlowUnits['CMD']
    <FlowUnits.CMD: (9, 1.1574074074074073e-05)>

    Units can be checked for metric or US customary status using the ``is_traditional`` or
    ``is_metric`` options.

    >>> FlowUnits.GPM.is_traditional
    True
    >>> FlowUnits.GPM.is_metric
    False

    Conversion can be done using the `factor` attribute. For example, to convert 10 AFD to SI units,
    and to convert 10 MGD to MLD,

    >>> 10 * FlowUnits.AFD.factor
    0.14276410185185184
    >>> 10 * FlowUnits.MGD.factor / FlowUnits.MLD.factor
    37.85411784000001


    .. note::
        This Enum uses a value of 0 for one of its members, and therefore
        acts in a non-standard way when evaluating truth values. Use ``None`` / ``is None``
        to check for truth values for variables storing a FlowUnits.

    """
    CFS = (0, 0.0283168466)
    GPM = (1, (0.003785411784/60.0))
    MGD = (2, (1e6*0.003785411784/86400.0))
    IMGD = (3, (1e6*0.00454609/86400.0))
    AFD = (4, (1233.48184/86400.0))
    LPS = (5, 0.001)
    LPM = (6, (0.001/60.0))
    MLD = (7, (1e6*0.001/86400.0))
    CMH = (8, (1.0/3600.0))
    CMD = (9, (1.0/86400.0))
    SI = (11, 1.0)

    def __init__(self, EN_id, flow_factor):
        self._value2member_map_[EN_id] = self
        self._member_map_[self.name.lower()] = self

    def __int__(self):
        """Convert to an EPANET Toolkit enum number."""
        return int(self.value[0])

    def __str__(self):
        """Convert to a string for INP files."""
        return self.name

    @property
    def factor(self):
        """float: The conversion factor to convert units into SI units of :math:`m^3\,s^{-1}`.

        Letting values in the original units be :math:`v`, and the resulting values in SI units
        be :math:`s`, the conversion factor, :math:`f`, such that

        .. math::
            v f = s

        """
        return self.value[1]

    @property
    def is_traditional(self):
        """bool: True if flow unit is a US Customary (traditional) unit.

        Traditional units include CFS, GPM, MGD, IMGD and AFD.

        Examples
        --------
        >>> FlowUnits.MGD.is_traditional
        True
        >>> FlowUnits.MLD.is_traditional
        False
        >>> FlowUnits.SI.is_traditional
        False

        """
        return self in [FlowUnits.CFS, FlowUnits.GPM, FlowUnits.MGD, FlowUnits.IMGD, FlowUnits.AFD]

    @property
    def is_metric(self):
        """bool: True if flow unit is an SI Derived (metric) unit.

        Metric units include LPS, LPM, MLD, CMH, and CMD.
        This 'does not' include FlowUnits.SI itself, only 'derived' units.

        Examples
        --------
        >>> FlowUnits.MGD.is_metric
        False
        >>> FlowUnits.MLD.is_metric
        True
        >>> FlowUnits.SI.is_metric
        False

        """
        return self in [FlowUnits.LPS, FlowUnits.LPM, FlowUnits.MLD, FlowUnits.CMH, FlowUnits.CMD]


class MassUnits(enum.Enum):
    r"""Mass units used by EPANET, plus SI conversion factor.

    Mass units are defined in the EPANET INP file when the QUALITY option is
    set to a chemical. This is parsed to obtain the mass part of the concentration units,
    and is used to set this enumerated type.

    .. rubric:: Enum Members

    ============  ============================================
    :attr:`~mg`   miligrams; EPANET as "mg/L" or "mg/min"
    :attr:`~ug`   micrograms; EPANET as "ug/L" or "ug/min"
    :attr:`~g`    grams
    :attr:`~kg`   kilograms; WNTR standard
    ============  ============================================

    .. rubric:: Enum Member Attributes

    .. autosummary::
        factor

    """
    mg = (1, 0.000001)
    ug = (2, 0.000000001)
    g = (3, 0.001)
    kg = (4, 1.0)

    @property
    def factor(self):
        """float : The scaling factor to convert to kg."""
        return self.value[1]


class QualParam(enum.Enum):
    u"""EPANET water quality parameters conversion.

    These parameters are separated from the HydParam parameters because they are related to a
    logically separate model in EPANET, but also because conversion to SI units requires additional
    information, namely, the MassUnits that were specified in the EPANET input file. Additionally,
    the reaction coefficient conversions require information about the reaction order that was
    specified. See the `to_si` and `from_si` functions for details.

    .. rubric:: Enum Members

    ==========================  ================================================================
    :attr:`~Concentration`      General concentration parameter
    :attr:`~Quality`            Nodal water quality
    :attr:`~LinkQuality`        Link water quality
    :attr:`~BulkReactionCoeff`  Bulk reaction coefficient (req. `reaction_order` to convert)
    :attr:`~WallReactionCoeff`  Wall reaction coefficient (req. `reaction_order` to convert)
    :attr:`~ReactionRate`       Average reaction rate within a link
    :attr:`~SourceMassInject`   Injection rate for water quality sources
    :attr:`~WaterAge`           Water age at a node
    ==========================  ================================================================


    """
    Quality = 4
    LinkQuality = 10
    ReactionRate = 13
    Concentration = 35
    BulkReactionCoeff = 36
    WallReactionCoeff = 37
    SourceMassInject = 38
    WaterAge = 39

    def __init__(self, value):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def _to_si(self, flow_units, data, mass_units=MassUnits.mg,
              reaction_order=0):
        """Convert a water quality parameter to SI units from EPANET units.

        By default, the mass units are the EPANET default of mg, and the reaction order is 0.

        Parameters
        ----------
        flow_units : ~FlowUnits
            The EPANET flow units to use in the conversion
        data : array-like
            The data to be converted (scalar, array or dictionary)
        mass_units : ~MassUnits
            The EPANET mass units to use in the conversion (mg or ug)
        reaction_order : int
            The reaction order for use converting reaction coefficients

        Returns
        -------
        float
            The data values converted to SI standard units

        """
        data_type = type(data)
        if data_type is dict:
            data_keys = data.keys()
            data = np.array(data.values())
        elif data_type is list:
            data = np.array(data)

        if mass_units is None:
            mass_units=MassUnits.mg

        # Do conversions
        if self in [QualParam.Concentration, QualParam.Quality, QualParam.LinkQuality]:
            data = data * (mass_units.factor/0.001)  # MASS /L to kg/m3

        elif self in [QualParam.SourceMassInject]:
            data = data * (mass_units.factor/60.0)  # MASS /min to kg/s

        elif self in [QualParam.BulkReactionCoeff] and reaction_order == 1:
                data = data * (1/86400.0)  # per day to per second

        elif self in [QualParam.WallReactionCoeff] and reaction_order == 0:
            if flow_units.is_traditional:
                data = data * (mass_units.factor*0.092903/86400.0)  # M/ft2/d to SI
            else:
                data = data * (mass_units.factor/86400.0)  # M/m2/day to M/m2/s

        elif self in [QualParam.WallReactionCoeff] and reaction_order == 1:
            if flow_units.is_traditional:
                data = data * (0.3048/86400.0)  # ft/d to m/s
            else:
                data = data * (1.0/86400.0)  # m/day to m/s

        elif self in [QualParam.SourceMassInject]:
            data = data * (mass_units.factor/60.0)  # per min to per second

        elif self in [QualParam.WaterAge]:
            data = data * 3600.0  # hr to s

        # Convert back to input data type
        if data_type is dict:
            data = dict(zip(data_keys, data))
        elif data_type is list:
            data = list(data)
        return data

    def _from_si(self, flow_units, data, mass_units=MassUnits.mg,
                reaction_order=0):
        """Convert a water quality parameter back to EPANET units from SI units.

        Mass units defaults to :class:`MassUnits.mg`, as this is the EPANET default.

        Parameters
        ----------
        flow_units : ~FlowUnits
            The EPANET flow units to use in the conversion
        data : array-like
            The SI unit data to be converted (scalar, array or dictionary)
        mass_units : ~MassUnits
            The EPANET mass units to use in the conversion (mg or ug)
        reaction_order : int
            The reaction order for use converting reaction coefficients

        Returns
        -------
        float
            The data values converted to EPANET appropriate units, based on the flow units.

        """
        data_type = type(data)
        if data_type is dict:
            data_keys = data.keys()
            data = np.array(data.values())
        elif data_type is list:
            data = np.array(data)

        # Do conversions
        if self in [QualParam.Concentration, QualParam.Quality, QualParam.LinkQuality]:
            data = data / (mass_units.factor/0.001)  # MASS /L fr kg/m3

        elif self in [QualParam.SourceMassInject]:
            data = data / (mass_units.factor/60.0)  # MASS /min fr kg/s

        elif self in [QualParam.BulkReactionCoeff] and reaction_order == 1:
                data = data / (1/86400.0)  # per day fr per second

        elif self in [QualParam.WallReactionCoeff] and reaction_order == 0:
            if flow_units.is_traditional:
                data = data / (mass_units.factor*0.092903/86400.0)  # M/ft2/d fr SI
            else:
                data = data / (mass_units.factor/86400.0)  # M/m2/day fr M/m2/s

        elif self in [QualParam.WallReactionCoeff] and reaction_order == 1:
            if flow_units.is_traditional:
                data = data / (0.3048/86400.0)  # ft/d fr m/s
            else:
                data = data / (1.0/86400.0)  # m/day fr m/s

        elif self in [QualParam.SourceMassInject]:
            data = data / (mass_units.factor/60.0)  # per min fr per second

        elif self in [QualParam.WaterAge]:
            data = data / 3600.0  # hr fr s

        # Convert back to input data type
        if data_type is dict:
            data = dict(zip(data_keys, data))
        elif data_type is list:
            data = list(data)
        return data


class HydParam(enum.Enum):
    u"""EPANET hydraulics and energy parameter conversion.

    The hydraulic parameter enumerated type is used to perform unit conversion
    between EPANET internal units and SI units used by WNTR. The units for each
    parameter are determined based on the :class:`FlowUnits` used.

    Parameters that are unitless or otherwise require no conversion are not members of this
    Enum type.

    .. rubric:: Enum Members

    ==========================  ===================================================================
    :attr:`~Elevation`          Nodal elevation
    :attr:`~Demand`             Nodal demand
    :attr:`~HydraulicHead`      Nodal head
    :attr:`~Pressure`           Nodal pressure
    :attr:`~EmitterCoeff`       Emitter coefficient
    :attr:`~TankDiameter`       Tank diameter
    :attr:`~Volume`             Tank volume
    :attr:`~Length`             Link length
    :attr:`~PipeDiameter`       Pipe diameter
    :attr:`~Flow`               Link flow
    :attr:`~Velocity`           Link velocity
    :attr:`~HeadLoss`           Link headloss (from start node to end node)
    :attr:`~RoughnessCoeff`     Link roughness (requires `darcy_weisbach` setting for conversion)
    :attr:`~Energy`             Pump energy
    :attr:`~Power`              Pump power
    ==========================  ===================================================================


    """
    Elevation = 0
    Demand = 1
    HydraulicHead = 2
    Pressure = 3

    # Quality = 4

    Length = 5
    PipeDiameter = 6
    Flow = 7
    Velocity = 8
    HeadLoss = 9

    # Link Quality = 10
    # Status = 11
    # Setting = 12
    # Reaction Rate = 13
    # Friction factor = 14

    Power = 15

    # Time = 16

    Volume = 17

    # The following are not "output" network variables, and thus are defined separately

    EmitterCoeff = 31
    RoughnessCoeff = 32
    TankDiameter = 33
    Energy = 34

    def __init__(self, value):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def _to_si(self, flow_units, data, darcy_weisbach=False):
        """Convert from EPANET units groups to SI units.

        If converting roughness, specify if the Darcy-Weisbach equation is
        used using the darcy_weisbach parameter. Otherwise, that parameter
        can be safely ignored/omitted for any other conversion.

        Parameters
        ----------
        flow_units : ~FlowUnits
            The flow units to use in the conversion
        data : array-like
            The EPANET-units data to be converted (scalar, array or dictionary)
        darcy_weisbach : bool, optional
            Set to ``True`` if converting roughness coefficients for use with Darcy-Weisbach
            formula.

        Returns
        -------
        float
            The data values converted to SI standard units.

        """
        # Convert to array for unit conversion
        data_type = type(data)
        if data_type is pd.core.frame.DataFrame:
            data_index = data.index
            data_columns = data.columns
            data = data.values
        elif data_type is dict:
            data_keys = data.keys()
            data = np.array(list(data.values()))
        elif data_type is list:
            data = np.array(data)

        # Do conversions
        if self in [HydParam.Demand, HydParam.Flow, HydParam.EmitterCoeff]:
            data = data * flow_units.factor
            if self is HydParam.EmitterCoeff:
                if flow_units.is_traditional:
                    data = data * np.sqrt(0.7032)  # flowunit/psi0.5 to flowunit/m0.5

        elif self in [HydParam.PipeDiameter]:
            if flow_units.is_traditional:
                data = data * 0.0254  # in to m
            elif flow_units.is_metric:
                data = data * 0.001  # mm to m

        elif self in [HydParam.RoughnessCoeff] and darcy_weisbach:
            if flow_units.is_traditional:
                data = data * (1000.0*0.3048)  # 1e-3 ft to m
            elif flow_units.is_metric:
                data = data * 0.001  # mm to m

        elif self in [HydParam.TankDiameter, HydParam.Elevation, HydParam.HydraulicHead,
                      HydParam.Length]:#, HydParam.HeadLoss]:
            if flow_units.is_traditional:
                data = data * 0.3048  # ft to m

        elif self in [HydParam.Velocity]:
            if flow_units.is_traditional:
                data = data * 0.3048  # ft/s to m/s

        elif self in [HydParam.Energy]:
            data = data * 3600000.0  # kW*hr to J

        elif self in [HydParam.Power]:
            if flow_units.is_traditional:
                data = data * 745.699872  # hp to W (Nm/s)
            elif flow_units.is_metric:
                data = data * 1000.0  # kW to W (Nm/s)

        elif self in [HydParam.Pressure]:
            if flow_units.is_traditional:
#                data = data * 0.703249614902
                data = data * (0.3048/0.4333)  # psi * (m/ft / psi/ft)

        elif self in [HydParam.Volume]:
            if flow_units.is_traditional:
                data = data * np.power(0.3048, 3)  # ft3 to m3

        # Convert back to input data type
        if data_type is pd.core.frame.DataFrame:
            data = pd.DataFrame(data, columns=data_columns, index=data_index)
        elif data_type is dict:
            data = dict(zip(data_keys, data))
        elif data_type is list:
            data = list(data)
        return data

    def _from_si(self, flow_units, data, darcy_weisbach=False):
        """Convert from SI units into EPANET specified units.

        If converting roughness, specify if the Darcy-Weisbach equation is
        used using the darcy_weisbach parameter. Otherwise, that parameter
        can be safely ignored/omitted for any other conversion.

        Parameters
        ----------
        flow_units : :class:`~FlowUnits`
            The flow units to use in the conversion
        data : array-like
            The SI unit data to be converted (scalar, array or dictionary)
        darcy_weisbach : bool, optional
            Set to ``True`` if converting roughness coefficients for use with Darcy-Weisbach
            formula.

        Returns
        -------
        float
            The data values converted to EPANET appropriate units based on the flow units.

        """
        # Convert to array for conversion
        data_type = type(data)
        if data_type is dict:
            data_keys = data.keys()
            data = np.array(list(data.values()))
        elif data_type is list:
            data = np.array(data)

        # Do onversions
        if self in [HydParam.Demand, HydParam.Flow, HydParam.EmitterCoeff]:
            data = data / flow_units.factor
            if self is HydParam.EmitterCoeff:
                if flow_units.is_traditional:
                    data = data / np.sqrt(0.7032)  # flowunit/psi0.5 from flowunit/m0.5

        elif self in [HydParam.PipeDiameter]:
            if flow_units.is_traditional:
                data = data / 0.0254  # in from m
            elif flow_units.is_metric:
                data = data / 0.001  # mm from m

        elif self in [HydParam.RoughnessCoeff] and darcy_weisbach:
            if flow_units.is_traditional:
                data = data / (1000.0*0.3048)  # 1e-3 ft from m
            elif flow_units.is_metric:
                data = data / 0.001  # mm from m

        elif self in [HydParam.TankDiameter, HydParam.Elevation, HydParam.HydraulicHead,
                      HydParam.Length]:#, HydParam.HeadLoss]:
            if flow_units.is_traditional:
                data = data / 0.3048  # ft from m

        elif self in [HydParam.Velocity]:
            if flow_units.is_traditional:
                data = data / 0.3048  # ft/s from m/s

        elif self in [HydParam.Energy]:
            data = data / 3600000.0  # kW*hr from J

        elif self in [HydParam.Power]:
            if flow_units.is_traditional:
                data = data / 745.699872  # hp from W (Nm/s)
            elif flow_units.is_metric:
                data = data / 1000.0  # kW from W (Nm/s)

        elif self in [HydParam.Pressure]:
            if flow_units.is_traditional:
#                data = data / 0.703249614902
                data = data / (0.3048/0.4333)  # psi from mH2O, i.e. psi / (m/ft / psi/ft) )

        elif self in [HydParam.Volume]:
            if flow_units.is_traditional:
                data = data / np.power(0.3048, 3)  # ft3 from m3

        # Put back into data format passed in
        if data_type is dict:
            data = dict(zip(data_keys, data))
        elif data_type is list:
            data = list(data)
        return data


def to_si(from_units, data, param,
          mass_units=MassUnits.mg, pressure_units=None,
          darcy_weisbach=False, reaction_order=0):
    """Convert an EPANET parameter from internal to SI standard units.

    .. note:: 

        See the `Units <../units.html>`__ page for details on the units for each :class:`~HydParam` or :class:`~QualParam`.
        Other than for flows, most parameters only have one US and one metric unit that is used by EPANET.
        For example, even though flow might be specified in gallons, volumes would be specified in cubic feet
        for any US/English flow rates, and in cubic meters for all metric flow units; e.g., never liters 
        for volumes even when flow is declared as LPS.

        Rememeber that internally, WNTR is **always** expecting the values for a parameter to be in true SI
        units -- meters, kilograms, and seconds -- unless explicitly stated otherwise (e.g., hours for control times).


    Parameters
    ----------
    from_units : :class:`~FlowUnits`
        The EPANET flow units (and therefore units system) to use for conversion
    data : float, array-like, dict
        The data to be converted
    param : :class:`~HydParam` or :class:`~QualParam`
        The parameter type for the data
    mass_units : :class:`~MassUnits`, optional
        The EPANET mass units (mg or ug internal to EPANET)
    pressure_units : :class:`~PressureUnits`, optional
        The EPANET pressure units being used (based on `flow_units`, normally)
    darcy_weisbach : bool, optional
        For roughness coefficients, is this used in a Darcy-Weisbach formula?
    reaction_order : int, optional
        For reaction coefficients, what is the reaction order?

    Returns
    -------
    float, array-like, or dict
        The data values convert into SI standard units


    Examples
    --------

    First, we convert an array of flows from GPM to cubic meters per second (the SI units).

    >>> from wntr.epanet.util import *
    >>> flow_si = to_si(FlowUnits.GPM, [0.1, 1.0, 4.3], HydParam.Flow)
    >>> print(flow_si)
    [6.309019640000001e-06, 6.30901964e-05, 0.00027128784452]

    Next, we show how to convert the quality parameter from the EPANET units of mg/L to kg/m3.
    If that is not the mass units you prefer, it is possible to change them to ug/L, g/L, or kg/L,
    as shown in the second example.

    >>> to_si(FlowUnits.GPM, 4.6, QualParam.Quality)
    0.0046
    >>> to_si(FlowUnits.GPM, 4.6, QualParam.Quality, mass_units=MassUnits.ug)
    4.599999999999999e-06

    It is also possible to convert a dictionary of values.

    >>> to_si(FlowUnits.GPM, {'node1': 5.6, 'node2': 1.2}, HydParam.Pressure)
    {'node1': 3.9392568659127623, 'node2': 0.8441264712670206}

    For certain coefficients, there are flags that will change how the conversion occurs. For example,
    reaction coefficients depend on the reaction order.

    >>> to_si(FlowUnits.GPM, 0.45, QualParam.BulkReactionCoeff, reaction_order=0)
    0.45
    >>> to_si(FlowUnits.GPM, 0.45, QualParam.BulkReactionCoeff, reaction_order=1)
    5.208333333333333e-06


    """
    if isinstance(param, HydParam):
        return param._to_si(from_units, data, darcy_weisbach)
    elif isinstance(param, QualParam):
        return param._to_si(from_units, data, mass_units, reaction_order)
    else:
        raise RuntimeError('Invalid parameter: %s' % param)


def from_si(to_units, data, param,
          mass_units=MassUnits.mg, pressure_units=None,
          darcy_weisbach=False, reaction_order=0):
    """Convert an EPANET parameter from SI standard units back to internal units.

    .. note:: 

        See the `Units <../units.html>`__ page for details on the units for each :class:`~HydParam` or :class:`~QualParam`.
        Other than for flows, most parameters only have one US and one metric unit that is used by EPANET.
        For example, even though flow might be specified in gallons, volumes would be specified in cubic feet
        for any US/English flow rates, and in cubic meters for all metric flow units; e.g., never liters 
        for volumes even when flow is declared as LPS.

        Rememeber that internally, WNTR is **always** expecting the values for a parameter to be in true SI
        units -- meters, kilograms, and seconds -- unless explicitly stated otherwise (e.g., hours for control times).


    Parameters
    ----------
    to_units : :class:`~FlowUnits`
        The EPANET flow units (and therefore units system) to use for conversion
    data : float, array-like, dict
        The data to be converted
    param : :class:`~HydParam` or :class:`~QualParam`
        The parameter type for the data
    mass_units : :class:`~MassUnits`, optional
        The EPANET mass units (mg or ug internal to EPANET)
    pressure_units : :class:`~PressureUnits`, optional
        The EPANET pressure units being used (based on `flow_units`, normally)
    darcy_weisbach : bool, optional
        For roughness coefficients, is this used in a Darcy-Weisbach formula?
    reaction_order : int, optional
        For reaction coefficients, what is the reaction order?

    Returns
    -------
    float, array-like, or dict
        The data values converted into EPANET internal units


    Examples
    --------

    First, we convert an array of flows from SI (cubic meters per second) to GPM.

    >>> from wntr.epanet.util import *
    >>> flow_us = from_si(FlowUnits.GPM, [6.309019640000001e-06, 6.30901964e-05, 0.00027128784452], HydParam.Flow)
    >>> print(flow_us)
    [0.1, 1.0, 4.3]

    Next, we show how to convert the quality parameter from kg/m3 to mg/L and then to ug/L.
    
    >>> from_si(FlowUnits.GPM, 0.0046, QualParam.Quality)
    4.6
    >>> from_si(FlowUnits.GPM, 0.0046, QualParam.Quality, mass_units=MassUnits.ug)
    4600.0

    It is also possible to convert a dictionary of values.

    >>> from_si(FlowUnits.GPM, {'node1': 3.9392568659127623, 'node2': 0.8441264712670206}, HydParam.Pressure)
    {'node1': 5.6, 'node2': 1.2}
    
    Finally, an example showing the conversion of 1000 cubic meters per second into the different flow units.

    >>> from_si(FlowUnits.GPM, 1000.0, HydParam.Flow)  # to gallons per minute
    15850323.141488904
    >>> from_si(FlowUnits.LPS, 1000.0, HydParam.Flow)  # to liters per second
    1000000.0
    >>> from_si(FlowUnits.MGD, 1000.0, HydParam.Flow)  # to million gallons per day
    22824.465323744018


    """
    if isinstance(param, HydParam):
        return param._from_si(to_units, data, darcy_weisbach)
    elif isinstance(param, QualParam):
        return param._from_si(to_units, data, mass_units, reaction_order)
    else:
        raise RuntimeError('Invalid parameter: %s' % param)


class StatisticsType(enum.Enum):
    """EPANET time series statistics processing.

    .. rubric:: Enum Members

    ================  =========================================================================
    :attr:`~none`     Do no processing, provide instantaneous values on output at time `t`.
    :attr:`~Average`  Average the value across the report period ending at time `t`.
    :attr:`~Minimum`  Provide the minimum value across all complete reporting periods.
    :attr:`~Maximum`  Provide the maximum value across all complete reporting periods.
    :attr:`~Range`    Provide the range (max - min) across all complete reporting periods.
    ================  =========================================================================

    """
    none = 0
    Average = 1
    Minimum = 2
    Maximum = 3
    Range = 4

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name


class QualType(enum.Enum):
    """Provide the EPANET water quality simulation mode.

    .. rubric:: Enum Members

    ================  =========================================================================
    :attr:`~none`     Do not perform water quality simulation.
    :attr:`~Chem`     Do chemical transport simulation.
    :attr:`~Age`      Do water age simulation.
    :attr:`~Trace`    Do a tracer test (results in percentage of water is from trace node).
    ================  =========================================================================

    """
    none = 0
    Chem = 1
    Age = 2
    Trace = 3

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name


class SourceType(enum.Enum):
    """What type of EPANET Chemical source is used.

    .. rubric:: Enum Members

    ==================  =========================================================================
    :attr:`~Concen`     Concentration -- cannot be used at nodes with non-zero demand.
    :attr:`~Mass`       Mass -- mass per minute injection. Can be used at any node.
    :attr:`~Setpoint`   Setpoint -- force node quality to be a certain concentration.
    :attr:`~FlowPaced`  Flow paced -- set variable mass injection based on flow.
    ==================  =========================================================================

    """
    Concen = 0
    Mass = 1
    Setpoint = 2
    FlowPaced = 3

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name


class PressureUnits(enum.Enum):
    """EPANET output pressure units.

    .. rubric:: Enum Members

    ===============  ====================================================
    :attr:`~psi`     Pounds per square inch (flow units are traditional)
    :attr:`~kPa`     kilopascals (flow units are metric)
    :attr:`~meters`  meters of H2O
    ===============  ====================================================

    """
    psi = 0
    kPa = 1
    Meters = 2

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name


class FormulaType(enum.Enum):
    """Formula used for determining head loss due to roughness.

    .. rubric:: Enum Members

    ===============  ==================================================================
    :attr:`~HW`      Hazen-Williams headloss formula
    :attr:`~DW`      Darcy-Weisbach formala; requires units conversion
    :attr:`~CM`      Chezy-Manning formula
    ===============  ==================================================================

    """
    HW = (0, "H-W",)
    DW = (1, "D-W",)
    CM = (2, "C-M",)

    def __init__(self, eid, inpcode):
        self._value2member_map_[eid] = self
        self._member_map_[inpcode] = self
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __int__(self):
        return self.value[0]

    def __str__(self):
        return self.value[1]


class ControlType(enum.Enum):
    """The type of control.

    .. rubric:: Enum Members

    ==================  ==================================================================
    :attr:`~LowLevel`   Act when grade below set level
    :attr:`~HiLevel`    Act when grade above set level
    :attr:`~Timer`      Act when set time reached (from start of simulation)
    :attr:`~TimeOfDay`  Act when time of day occurs (each day)
    ==================  ==================================================================

    """

    LowLevel = 0
    HiLevel = 1
    Timer = 2
    TimeOfDay = 3

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name


class LinkTankStatus(enum.Enum):
    """The link tank status.

    .. rubric:: Enum Members

    ====================  ==================================================================
    :attr:`~XHead`        Pump cannot deliver head (closed)
    :attr:`~TempClosed`   Temporarily closed
    :attr:`~Closed`       Closed
    :attr:`~Open`         Open
    :attr:`~Active`       Valve active (partially open)
    :attr:`~XFlow`        Pump exceeds maximum flow
    :attr:`~XFCV`         FCV cannot supply flow
    :attr:`~XPressure`    Valve cannot supply pressure
    :attr:`~Filling`      Tank filling
    :attr:`~Emptying`     Tank emptying
    ====================  ==================================================================

    """
    
    XHead = 0  
    TempClosed = 1  
    Closed = 2 
    Open = 3 
    Active = 4  
    XFlow = 5  
    XFCV = 6  
    XPressure = 7  
    Filling = 8  
    Emptying = 9 

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name


class MixType(enum.Enum):
    """Tank mixing model type.

    .. rubric:: Enum Members

    ===============  ==================================================================
    :attr:`~Mix1`    Single compartment mixing model
    :attr:`~Mix2`    Two-compartment mixing model
    :attr:`~FIFO`    First-in/first-out model
    :attr:`~LIFO`    Last-in/first-out model
    ===============  ==================================================================

    """
    Mix1 = 0
    Mix2 = 1
    FIFO = 2
    LIFO = 3
    Mixed = 0
    TwoComp = 1

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name


class ResultType(enum.Enum):
    """Extended period simulation results type"""
    demand = 1
    head = 2
    pressure = 3
    quality = 4
    flowrate = 5
    velocity = 6
    headloss = 7
    linkquality = 8
    status = 9
    setting = 10
    rxnrate = 11
    frictionfact = 12

    @property
    def is_node(self):
        """Is a nodal property result"""
        if abs(self.value) < 5:
            return True
        return False

    @property
    def is_link(self):
        """Is a link property result"""
        if self.value > 4:
            return True
        return False

    @property
    def is_qual(self):
        """Is related to quality"""
        if self.value in [4, 8, 11]:
            return True
        return False

    @property
    def is_hyd(self):
        """Is related to hydraulics"""
        if self.value in [1,2,3,5,6,7,12]:
            return True
        return False


class EN(enum.IntEnum):
    """All the ``EN_`` constants for the EPANET toolkit.

    For example, ``EN_LENGTH`` is accessed as ``EN.LENGTH``, instead.  Please see the EPANET
    toolkit documentation for the description of these enums. Several enums are duplicated
    in separate classes above for clarity during programming.

    The enums can be broken in the following groups.

    - Node parameters: :attr:`~ELEVATION`, :attr:`~BASEDEMAND`, :attr:`~PATTERN`, :attr:`~EMITTER`, :attr:`~INITQUAL`, :attr:`~SOURCEQUAL`, :attr:`~SOURCEPAT`, :attr:`~SOURCETYPE`, :attr:`~TANKLEVEL`, :attr:`~DEMAND`, :attr:`~HEAD`, :attr:`~PRESSURE`, :attr:`~QUALITY`, :attr:`~SOURCEMASS`, :attr:`~INITVOLUME`, :attr:`~MIXMODEL`, :attr:`~MIXZONEVOL`, :attr:`~TANKDIAM`, :attr:`~MINVOLUME`, :attr:`~VOLCURVE`, :attr:`~MINLEVEL,`, :attr:`~MAXLEVEL`, :attr:`~MIXFRACTION`, :attr:`~TANK_KBULK`, :attr:`~TANKVOLUME`, :attr:`~MAXVOLUME`
    - Link parameters: :attr:`~DIAMETER`, :attr:`~LENGTH`, :attr:`~ROUGHNESS`, :attr:`~MINORLOSS`, :attr:`~INITSTATUS`, :attr:`~INITSETTING`, :attr:`~KBULK`, :attr:`~KWALL`, :attr:`~FLOW`, :attr:`~VELOCITY`, :attr:`~HEADLOSS`, :attr:`~STATUS`, :attr:`~SETTING`, :attr:`~ENERGY`, :attr:`~LINKQUAL`, :attr:`~LINKPATTERN`
    - Time parameters: :attr:`~DURATION`, :attr:`~HYDSTEP`, :attr:`~QUALSTEP`, :attr:`~PATTERNSTEP`, :attr:`~PATTERNSTART`, :attr:`~REPORTSTEP`, :attr:`~REPORTSTART`, :attr:`~RULESTEP`, :attr:`~STATISTIC`, :attr:`~PERIODS`, :attr:`~STARTTIME`, :attr:`~HTIME`, :attr:`~HALTFLAG`, :attr:`~NEXTEVENT`
    - Solver parameters: :attr:`~ITERATIONS`, :attr:`~RELATIVEERROR`
    - Component counts: :attr:`~NODECOUNT`, :attr:`~TANKCOUNT`, :attr:`~LINKCOUNT`, :attr:`~PATCOUNT`, :attr:`~CURVECOUNT`, :attr:`~CONTROLCOUNT`
    - Node types: :attr:`~JUNCTION`, :attr:`~RESERVOIR`, :attr:`~TANK`
    - Link types: :attr:`~CVPIPE`, :attr:`~PIPE`, :attr:`~PUMP`, :attr:`~PRV`, :attr:`~PSV`, :attr:`~PBV`, :attr:`~FCV`, :attr:`~TCV`, :attr:`~GPV`
    - Quality analysis types: :attr:`~NONE`, :attr:`~CHEM`, :attr:`~AGE`, :attr:`~TRACE`
    - Source quality types: :attr:`~CONCEN`, :attr:`~MASS`, :attr:`~SETPOINT`, :attr:`~FLOWPACED`
    - Flow unit types: :attr:`~CFS`, :attr:`~GPM`, :attr:`~MGD`, :attr:`~IMGD`, :attr:`~AFD`, :attr:`~LPS`, :attr:`~LPM`, :attr:`~MLD`, :attr:`~CMH`, :attr:`~CMD`
    - Miscelaneous options: :attr:`~TRIALS`, :attr:`~ACCURACY`, :attr:`~TOLERANCE`, :attr:`~EMITEXPON`, :attr:`~DEMANDMULT`
    - Control types: :attr:`~LOWLEVEL`, :attr:`~HILEVEL`, :attr:`~TIMER`, :attr:`~TIMEOFDAY`
    - Time statistic types: :attr:`~NONE`, :attr:`~AVERAGE`, :attr:`~MINIMUM`, :attr:`~MAXIMUM`, :attr:`~RANGE`
    - Tank mixing model types: :attr:`~MIX1`, :attr:`~MIX2`, :attr:`~FIFO`, :attr:`~LIFO`
    - Save results flag: :attr:`~NOSAVE`, :attr:`~SAVE`, :attr:`~INITFLOW`
    - Pump behavior types: :attr:`~CONST_HP`, :attr:`~POWER_FUNC`, :attr:`~CUSTOM`


    """
    # Node parameters
    ELEVATION    = 0
    BASEDEMAND   = 1
    PATTERN      = 2
    EMITTER      = 3
    INITQUAL     = 4
    SOURCEQUAL   = 5
    SOURCEPAT    = 6
    SOURCETYPE   = 7
    TANKLEVEL    = 8
    DEMAND       = 9
    HEAD         = 10
    PRESSURE     = 11
    QUALITY      = 12
    SOURCEMASS   = 13
    INITVOLUME   = 14
    MIXMODEL     = 15
    MIXZONEVOL   = 16
    TANKDIAM     = 17
    MINVOLUME    = 18
    VOLCURVE     = 19
    MINLEVEL     = 20
    MAXLEVEL     = 21
    MIXFRACTION  = 22
    TANK_KBULK   = 23
    TANKVOLUME   = 24
    MAXVOLUME    = 25

    # Link parameters
    DIAMETER     = 0
    LENGTH       = 1
    ROUGHNESS    = 2
    MINORLOSS    = 3
    INITSTATUS   = 4
    INITSETTING  = 5
    KBULK        = 6
    KWALL        = 7
    FLOW         = 8
    VELOCITY     = 9
    HEADLOSS     = 10
    STATUS       = 11
    SETTING      = 12
    ENERGY       = 13
    LINKQUAL     = 14
    LINKPATTERN  = 15

    # Time parameters
    DURATION     = 0
    HYDSTEP      = 1
    QUALSTEP     = 2
    PATTERNSTEP  = 3
    PATTERNSTART = 4
    REPORTSTEP   = 5
    REPORTSTART  = 6
    RULESTEP     = 7
    STATISTIC    = 8
    PERIODS      = 9
    STARTTIME    = 10
    HTIME        = 11
    HALTFLAG     = 12
    NEXTEVENT    = 13

    # Solver parameters
    ITERATIONS   = 0
    RELATIVEERROR= 1

    # Count parameters
    NODECOUNT    = 0
    TANKCOUNT    = 1
    LINKCOUNT    = 2
    PATCOUNT     = 3
    CURVECOUNT   = 4
    CONTROLCOUNT = 5

    # Node Types
    JUNCTION     = 0
    RESERVOIR    = 1
    TANK         = 2

    # Link Types
    CVPIPE       = 0
    PIPE         = 1
    PUMP         = 2
    PRV          = 3
    PSV          = 4
    PBV          = 5
    FCV          = 6
    TCV          = 7
    GPV          = 8

    # Quality Types
    NONE         = 0
    CHEM         = 1
    AGE          = 2
    TRACE        = 3

    # Source quality types
    CONCEN       = 0
    MASS         = 1
    SETPOINT     = 2
    FLOWPACED    = 3

    # Flow units parameter
    CFS          = 0
    GPM          = 1
    MGD          = 2
    IMGD         = 3
    AFD          = 4
    LPS          = 5
    LPM          = 6
    MLD          = 7
    CMH          = 8
    CMD          = 9

    # Miscelaneous parameters
    TRIALS       = 0
    ACCURACY     = 1
    TOLERANCE    = 2
    EMITEXPON    = 3
    DEMANDMULT   = 4

    # Control types
    LOWLEVEL     = 0
    HILEVEL      = 1
    TIMER        = 2
    TIMEOFDAY    = 3

    # Statistic Types
    AVERAGE      = 1
    MINIMUM      = 2
    MAXIMUM      = 3
    RANGE        = 4

    # Tank mixing parameters
    MIX1         = 0
    MIX2         = 1
    FIFO         = 2
    LIFO         = 3

    # Hydraulic solver / file parameters
    NOSAVE       = 0
    SAVE         = 1
    INITFLOW     = 10

    # Pump behavior Types
    CONST_HP     = 0
    POWER_FUNC   = 1
    CUSTOM       = 2



"""
The wntr.network.options module includes simulation options.

.. note:: 

    This module has been changed in version 0.2.3 to incorporate the new options 
    that EPANET 2.2 requires. It also reorganizes certain options to better align 
    with EPANET nomenclature. This change is not backwards compatible, particularly 
    when trying to use pickle files with older options.

.. rubric:: Contents

.. autosummary::

    Options
    TimeOptions
    GraphicsOptions
    HydraulicOptions
    ReportOptions
    ReactionOptions
    QualityOptions
    EnergyOptions
    UserOptions

"""
import re
import logging
import copy

logger = logging.getLogger(__name__)

def _float_or_None(value):
    """Converts a value to a float, but doesn't crash for values of None"""
    if value is not None:
        return float(value)
    return None


def _int_or_None(value):
    """Converts a value to an int, but doesn't crash for values of None"""
    if value is not None:
        return int(value)
    return None


def _new_report_params():
    ret = dict(elevation=False, demand=True, head=True, pressure=True,
                quality=True, length=False, diameter=False, flow=True,
                velocity=True, headloss=True, position=False, setting=False, reaction=False)
    ret['f-factor'] = False
    return ret


def _new_report_obj():
    ret = dict(demand=True, head=True, pressure=True, quality=True,
                flow=True, linkquality=True, velocity=True, headloss=True, status=True,
                setting=True, rxnrate=True, frictionfact=True)
    return ret


def _new_param_opts():
    ret = dict(elevation=dict(), demand=dict(), head=dict(), pressure=dict(),
                quality=dict(), length=dict(), diameter=dict(), flow=dict(), 
                velocity=dict(), headloss=dict(), position=dict(), setting=dict(),
                reaction=dict())
    ret['f-factor'] = dict()
    return ret


class _OptionsBase(object):
    @classmethod
    def factory(cls, val):
        """Create an options object based on passing in an instance of the object, a dict, or a tuple"""
        if isinstance(val, cls):
            return val
        elif isinstance(val, dict):
            return cls(**val)
        elif isinstance(val, (list, tuple)):
            return cls(*val)
        elif val is None:
            return cls()
        raise ValueError('Unknown type for %s.factory: %s',
                         cls.__name__, type(val))

    def __str__(self):
        return "{}({})".format(self.__class__.__name__, ", ".join(["{}={}".format(k, repr(v)) for k, v in self.__dict__.items()]))
    __repr__ = __str__

    def __iter__(self):
        for k, v in self.__dict__.items():
            try: 
                vv = dict(v)
            except:
                vv = v
            yield k, vv

    def __getitem__(self, index):
        return self.__dict__[index]
    
    def __eq__(self, other):
        if other is None: return False
        if not hasattr(other, '__dict__'): return False
        for k in self.__dict__.keys():
            if not self.__dict__[k] == other.__dict__[k]: return False
        return True


class TimeOptions(_OptionsBase):
    """
    Options related to simulation and model timing.
    These options are named according to the EPANET 2.2 "Times" settings.
    
    Parameters
    ----------
    duration : int
        Simulation duration (seconds). Defaults to 0.

    hydraulic_timestep : int
        Hydraulic timestep (seconds). Defaults to 3600 (one hour).

    quality_timestep : int
        Water quality timestep (seconds). Defaults to 360 (five minutes).

    rule_timestep : int
        Rule timestep (seconds). Defaults to 360 (five minutes).

    pattern_timestep : int
        Pattern timestep (seconds). Defaults to 3600 (one hour).

    pattern_start : int
        Time offset (in seconds) to find the starting pattern step; changes 
        where in pattern the pattern starts out, *not* what time the pattern 
        starts. Defaults to 0.

    report_timestep : int
        Reporting timestep (seconds). Defaults to 3600 (one hour).

    report_start : int
        Start time of the report (in seconds) from the start of the simulation. Default 0.

    start_clocktime : int
        Time of day (in seconds from midnight) at which the simulation begins. Default 0 (midnight).

    statistic: str
        Provide statistics rather than time series report in the report file.
        Options are AVERAGED, MINIMUM, MAXIUM, RANGE, and NONE (as defined in the 
        EPANET User Manual). Defaults to NONE.
    

    """
    _pattern1 = re.compile(r'^(\d+):(\d+):(\d+)$')
    _pattern2 = re.compile(r'^(\d+):(\d+)$')
    _pattern3 = re.compile(r'^(\d+)$')
    def __init__(self,
                duration: float = 0.0,
                hydraulic_timestep: float=3600.0,
                quality_timestep: float=360.0,
                rule_timestep: float=360.0,
                pattern_timestep: float=3600.0,
                pattern_start: float=0.0,
                report_timestep: float=3600.0,
                report_start: float=0.0,
                start_clocktime: float=0.0,
                statistic: str='NONE'):
        self.duration = duration
        self.hydraulic_timestep = hydraulic_timestep
        self.quality_timestep = quality_timestep
        self.rule_timestep = rule_timestep
        self.pattern_timestep = pattern_timestep
        self.pattern_start = pattern_start
        self.report_timestep = report_timestep
        self.report_start = report_start
        self.start_clocktime = start_clocktime
        self.statistic = statistic

    def __setattr__(self, name, value):
        if name == 'statistic':
            value = str.upper(value)
            if value not in ['AVERAGED', 'MINIMUM', 'MAXIMUM', 'RANGE', 'NONE']:
                raise ValueError('Statistic must be one of AVERAGED, MINIMUM, MAXIMUM, RANGE or NONE')
        elif name not in ['report_timestep']:
            try:
                value = float(value)
            except ValueError:
                raise ValueError('%s must be a number', name)
        self.__dict__[name] = value

    @classmethod
    def seconds_to_tuple(cls, sec):
        hours = int(sec/3600.)
        sec -= hours*3600
        mm = int(sec/60.)
        sec -= mm*60
        return (hours, mm, int(sec))

    @classmethod
    def time_str_to_seconds(cls, s):
        """
        Converts time format to seconds.

        Parameters
        ----------
        s : string
            Time string. Options are 'HH:MM:SS', 'HH:MM', 'HH'


        Returns
        -------
        Integer value of time in seconds.
        """
        time_tuple = cls._pattern1.search(s)
        if bool(time_tuple):
            return (int(time_tuple.groups()[0])*60*60 +
                    int(time_tuple.groups()[1])*60 +
                    int(round(float(time_tuple.groups()[2]))))
        else:
            time_tuple = cls._pattern2.search(s)
            if bool(time_tuple):
                return (int(time_tuple.groups()[0])*60*60 +
                        int(time_tuple.groups()[1])*60)
            else:
                time_tuple = cls._pattern3.search(s)
                if bool(time_tuple):
                    return int(time_tuple.groups()[0])*60*60
                else:
                    raise RuntimeError("Time format not recognized. ")

    @classmethod
    def clock_str_to_seconds(cls, s, am_pm):
        """
        Converts clocktime format to seconds.


        Parameters
        ----------
        s : string
            Time string. Options are 'HH:MM:SS', 'HH:MM', HH'

        am : string
            options are AM or PM


        Returns
        -------
        Integer value of time in seconds

        """
        if am_pm.upper() == 'AM':
            am = True
        elif am_pm.upper() == 'PM':
            am = False
        else:
            raise RuntimeError('am_pm option not recognized; options are AM or PM')

        time_tuple = cls._pattern1.search(s)
        if bool(time_tuple):
            time_sec = (int(time_tuple.groups()[0])*60*60 +
                        int(time_tuple.groups()[1])*60 +
                        int(round(float(time_tuple.groups()[2]))))
            if s.startswith('12'):
                time_sec -= 3600*12
            if not am:
                if time_sec >= 3600*12:
                    raise RuntimeError('Cannot specify am/pm for times greater than 12:00:00')
                time_sec += 3600*12
            return time_sec
        else:
            time_tuple = cls._pattern2.search(s)
            if bool(time_tuple):
                time_sec = (int(time_tuple.groups()[0])*60*60 +
                            int(time_tuple.groups()[1])*60)
                if s.startswith('12'):
                    time_sec -= 3600*12
                if not am:
                    if time_sec >= 3600 * 12:
                        raise RuntimeError('Cannot specify am/pm for times greater than 12:00:00')
                    time_sec += 3600*12
                return time_sec
            else:
                time_tuple = cls._pattern3.search(s)
                if bool(time_tuple):
                    time_sec = int(time_tuple.groups()[0])*60*60
                    if s.startswith('12'):
                        time_sec -= 3600*12
                    if not am:
                        if time_sec >= 3600 * 12:
                            raise RuntimeError('Cannot specify am/pm for times greater than 12:00:00')
                        time_sec += 3600*12
                    return time_sec
                else:
                    raise RuntimeError("Time format not recognized. ")


class GraphicsOptions(_OptionsBase):
    """
    Options related to graphics. 
    
    May be used to contain custom, user defined values. 
    These options are taken from the EPANET "[BACKDROP]" section. 
    Additionally, the "MAP" option (`map_filename`), which identifies a file containing 
    node coordinates in the EPANET "[OPTIONS]" section, is also included here.
    
    Parameters
    ----------
    dimensions : 4-tuple or list
        Dimensions for backdrop image in the order (LLx, LLy, URx, URy). By default, 
        EPANET will make the image match the full extent of node coordinates (set to `None`).

    units : str
        Units for backdrop image dimensions. Must be one of FEET, METERS, DEGREES or NONE. 
        Default is NONE.

    offset : 2-tuple or list
        Offset for the network in order (X, Y). Default is None (no offset).

    image_filename : string
        Filename where image is located. Default is None.

    map_filename : string
        Filename used to store node coordinates in (node, x, y) format. This option is
        from the EPANET "[OPTIONS]" section. See note below.
    
    
    .. note::

        Because the format of the MAP file is uncertain, file will need to be processed
        by the user to assign coordinates to nodes, if desired. Remember that node 
        coordinates have impact *only* on graphics and *do not* impact simulation results.
        If the `map_filename` is not ``None``, then no [COORDINATES] will be written out
        to INP files (to save space, since the simulator does not use that section).
        This can be overwritten in the `write_inpfile` commands.
    

    """
    def __init__(self,
                 dimensions: list = None,
                 units: str = None,
                 offset: list = None,
                 image_filename: str = None,
                 map_filename: str = None):
        self.dimensions = dimensions
        self.units = units
        self.offset = offset
        self.image_filename = image_filename
        self.map_filename = map_filename

    def __setattr__(self, name, value):
        if name == 'units':
            value = str(value).upper()
            if value not in ['FEET','METERS','DEGREES','NONE']:
                raise ValueError('Backdrop units must be one of FEET, METERS, DEGREES, or NONE')
        self.__dict__[name] = value


class HydraulicOptions(_OptionsBase): 
    """
    Options related to hydraulic model.
    These options are named according to the settings in the EPANET "[OPTIONS]"
    section. Unless specified, these options are valid for both EPANET 2.0 and 2.2.
    
    Parameters
    ----------
    headloss : str
        Formula to use for computing head loss through a pipe. Options are H-W, 
        D-W, and C-M. Default is `H-W`.

    hydraulics : str
        Indicates if a hydraulics file should be read in or saved; options are 
        None, USE and SAVE. Defaults to ``None``.

    hydraulics_filename : str
        Filename to use if ``hydraulics = 'SAVE'``. Defaults to ``None``.

    viscosity : float
        Kinematic viscosity of the fluid. Defaults to 1.0.

    specific_gravity : float
        Specific gravity of the fluid. Defaults to 1.0.

    pattern : str
        Name of the default pattern for junction demands. By default,
        the default pattern is the pattern named "1". If this is set 
        to None (or if pattern "1" does not exist), then
        junctions with demands but without patterns will be held constant.

    demand_multiplier : float
        The demand multiplier adjusts the values of baseline demands for all 
        junctions. Defaults to 1.0.

    emitter_exponent : float
        The exponent used when computing flow from an emitter. Defaults to 0.5.

    minimum_pressure : float
        (EPANET 2.2 only) The global minimum nodal pressure. Defaults to 0.0.

    required_pressure: float
        (EPANET 2.2 only) The required nodal pressure. Defaults to 0.07 (m H2O)

    pressure_exponent: float
        (EPANET 2.2 only) The pressure exponent. Defaults to 0.5.

    trials : int
        Maximum number of trials used to solve network hydraulics. Defaults to 200.

    accuracy : float
        Convergence criteria for hydraulic solutions. Defaults to 0.001.

    headerror : float
        (EPANET 2.2 only) Augments the `accuracy` option by adjusting the head 
        error convergence limit. Defaults to 0 (off).

    flowchange : float
        (EPANET 2.2 only) Augments the `accuracy` option by adjusting the flow 
        change convergence limit. Defaults to 0 (off).

    unbalanced : str
        Indicate what happens if a hydraulic solution cannot be reached.  
        Options are STOP and CONTINUE. Defaults to STOP.

    unbalanced_value : int
        Number of additional trials if unbalanced = CONTINUE. Default is None.

    checkfreq : int
        Number of solution trials that pass between status checks. Default is 2.

    maxcheck : int
        Number of solution trials that pass between status check. Default is 10.

    damplimit : float
        Accuracy value at which solution damping begins. Default is 0 (no damping).

    demand_model : str
        Demand model for EPANET 2.2; acceptable values are DD and PDD. Default is DD.
        EPANET 2.0 only contains demand driven analysis, and will issue a warning 
        if this option is not set to DD.

    inpfile_units : str
        Units for the INP-file; options are CFS, GPM, MGD, IMGD, AFD, LPS, 
        LPM, MLD, CMH, and CMD. This **only** changes the units used in generating
        the INP file -- it has **no impact** on the units used in WNTR, which are 
        **always** SI units (m, kg, s).
    

    """
    def __init__(self,
                 headloss: str = 'H-W',
                 hydraulics: str = None,
                 hydraulics_filename: str = None,
                 viscosity: float = 1.0,
                 specific_gravity: float = 1.0,
                 pattern: str = '1',
                 demand_multiplier: float = 1.0,
                 demand_model: str = 'DD',
                 minimum_pressure: float = 0.0,
                 required_pressure: float = 0.07,  # EPANET 2.2 default
                 pressure_exponent: float = 0.5,
                 emitter_exponent: float = 0.5,
                 trials: int = 200,  # EPANET 2.2 increased the default from 40 to 200
                 accuracy: float = 0.001,
                 unbalanced: str = 'STOP',
                 unbalanced_value: int = None,
                 checkfreq: int = 2,
                 maxcheck: int = 10,
                 damplimit: int = 0,
                 headerror: float = 0,
                 flowchange: float = 0,
                 inpfile_units: str = 'GPM'):
        self.headloss = headloss
        self.hydraulics = hydraulics
        self.hydraulics_filename = hydraulics_filename
        self.viscosity = viscosity
        self.specific_gravity = specific_gravity
        self.pattern = pattern
        self.demand_multiplier = demand_multiplier
        self.demand_model = demand_model
        self.minimum_pressure = minimum_pressure
        self.required_pressure = required_pressure
        self.pressure_exponent = pressure_exponent
        self.emitter_exponent = emitter_exponent
        self.trials = trials
        self.accuracy = accuracy
        self.unbalanced = unbalanced
        self.unbalanced_value = unbalanced_value
        self.checkfreq = checkfreq
        self.maxcheck = maxcheck
        self.damplimit = damplimit
        self.headerror = headerror
        self.flowchange = flowchange
        self.inpfile_units = inpfile_units

    def __setattr__(self, name, value):
        if name == 'headloss':
            value = str.upper(value)
            if value not in ['H-W', 'D-W', 'C-M']:
                raise ValueError('headloss must be one of "H-W", "D-W", or "C-M"')
        elif name == 'hydraulics':
            if value is not None:
                value = str.upper(value)
                if value not in ['USE', 'SAVE']:
                    raise ValueError('hydraulics must be None (off) or one of "USE" or "SAVE"')
        elif name == 'demand_model':
            if value is not None:
                value = str.upper(value)
                if value not in ['DDA', 'DD', 'PDD', 'PDA']:
                    raise ValueError('demand_model must be None (off) or one of "DD", "DDA", "PDD", or "PDA"')
                if value == 'DD': value = 'DDA'
                if value == 'PDD': value = 'PDA'
        elif name == 'unbalanced':
            value = str.upper(value)
            if value not in ['STOP', 'CONTINUE']:
                raise ValueError('headloss must be either "STOP" or "CONTINUE"')
        elif name == 'inpfile_units' and isinstance(value, str):
            value = str.upper(value)
            if value not in ['CFS', 'GPM', 'MGD', 'IMGD', 'AFD', 'LPS', 'LPM', 'MLD', 'CMH', 'CMD']:
                raise ValueError('inpfile_units = "%s" is not a valid EPANET unit code', value)
        elif name == 'unbalanced_value':
            try:
                value = _int_or_None(value)
            except ValueError:
                raise ValueError('%s must be an int or None', name)
        elif name in ['trials', 'checkfreq', 'maxcheck']:
            try:
                value = int(value)
            except ValueError:
                raise ValueError('%s must be an integer', name)
        elif name not in ['pattern', 'hydraulics_filename', 'inpfile_units']:
            try:
                value = float(value)
            except ValueError:
                raise ValueError('%s must be a number', name)
        self.__dict__[name] = value


class ReactionOptions(_OptionsBase):
    """
    Options related to water quality reactions.
    From the EPANET "[REACTIONS]" options.
    
    Parameters
    ----------
    bulk_order : float
        Order of reaction occurring in the bulk fluid. Defaults to 1.0.

    wall_order : float
        Order of reaction occurring at the pipe wall; must be either 0 or 1. Defaults to 1.0.

    tank_order : float
        Order of reaction occurring in the tanks. Defaults to 1.0.

    bulk_coeff : float
        Global reaction coefficient for bulk fluid and tanks. Defaults to 0.0.

    wall_coeff : float
        Global reaction coefficient for pipe walls. Defaults 0.0.

    limiting_potential : float
        Specifies that reaction rates are proportional to the difference 
        between the current concentration and some limiting potential value, 
        Defaults to None (off).

    roughness_correl : float
        Makes all default pipe wall reaction coefficients related to pipe 
        roughness, according to functions as defined in EPANET. Defaults
        to None (off).
        

    .. note::

        Remember to use positive numbers for growth reaction coefficients and 
        negative numbers for decay coefficients. The time units for all reaction
        coefficients are in "per-second" and converted to/from EPANET units during I/O.

    """
    def __init__(self,
                 bulk_order: float = 1.0,
                 wall_order: float = 1.0,
                 tank_order: float = 1.0,
                 bulk_coeff: float = 0.0,
                 wall_coeff: float = 0.0,
                 limiting_potential: float = None,
                 roughness_correl: float = None):
        self.bulk_order = bulk_order
        self.wall_order = wall_order
        self.tank_order = tank_order
        self.bulk_coeff = bulk_coeff
        self.wall_coeff = wall_coeff
        self.limiting_potential = limiting_potential
        self.roughness_correl = roughness_correl

    def __setattr__(self, name, value):
        if name not in ['limiting_potential', 'roughness_correl']:
            try:
                value = float(value)
            except ValueError:
                raise ValueError('%s must be a number', name)
        else:
            try:
                value = _float_or_None(value)
            except ValueError:
                raise ValueError('%s must be a number or None', name)
        self.__dict__[name] = value


class QualityOptions(_OptionsBase):
    """
    Options related to water quality modeling. These options come from
    the "[OPTIONS]" section of an EPANET INP file.
    
    Parameters
    ----------
    parameter : str
        Type of water quality analysis.  Options are NONE, CHEMICAL, AGE, and 
        TRACE. Defaults to None.

    trace_node : str
        Trace node name if quality = TRACE. Defaults to None.

    chemical : str
        Chemical name for 'chemical' analysis. Defaults to "CHEMICAL" if appropriate.

    diffusivity : float
        Molecular diffusivity of the chemical. Defaults to 1.0.

    tolerance : float
        Water quality solver tolerance. Defaults to 0.01.

    inpfile_units : str
        Units for quality analysis if the parameter is set to CHEMICAL. 
        This is **only** relevant for the INP file. This value **must** be either
        "mg/L" (default) or "ug/L" (miligrams or micrograms per liter). 
        Internal WNTR units are always SI units (kg/m3).


    """
    def __init__(self,
                 parameter: str = 'NONE',
                 trace_node: str = None,
                 chemical_name: str = 'CHEMICAL',
                 diffusivity: float = 1.0,
                 tolerance: float = 0.01,
                 inpfile_units: str = 'mg/L'):
        self.parameter = parameter
        self.trace_node = trace_node
        self.chemical_name = chemical_name
        self.diffusivity = diffusivity
        self.tolerance = tolerance
        self.inpfile_units = inpfile_units

    def __setattr__(self, name, value):
        if name in ['diffusivity', 'tolerance']:
            try:
                value = float(value)
            except ValueError:
                raise ValueError('%s must be a number or None', name)
        self.__dict__[name] = value


class EnergyOptions(_OptionsBase):
    """
    Options related to energy calculations.
    From the EPANET "[ENERGY]" settings.
    
    Parameters
    ----------
    global_price : float
        Global average cost per Joule. Defaults to 0.

    global_pattern : str
        ID label of time pattern describing how energy price varies with time.
        Defaults to None.

    global_efficiency : float
        Global pump efficiency as percent; i.e., 75.0 means 75%.
        Defaults to None.

    demand_charge : float
        Added cost per maximum kW usage during the simulation period.
        Defaults to None.

    
    """
    def __init__(self,
                global_price: float=0,
                global_pattern: str=None,
                global_efficiency: float=None,
                demand_charge: float=None):
        self.global_price = global_price
        self.global_pattern = global_pattern
        self.global_efficiency = global_efficiency
        self.demand_charge = demand_charge


class ReportOptions(_OptionsBase):
    """
    Options related to EPANET report outputs. 
    The values in this options class *do not* affect the behavior of the WNTRSimulator.
    These only affect what is written to an EPANET INP file and the results that are
    in the EPANET-created report file.

    Parameters
    ----------
    report_filename : str
        Provides the filename to use for outputting an EPANET report file.
        By default, this will be the prefix plus ".rpt".

    status : str
        Output solver status ('YES', 'NO', 'FULL'). 'FULL' is only useful for debugging

    summary : str
        Output summary information ('YES' or 'NO')

    energy : str
        Output energy information

    nodes : None, "ALL", or list
        Output node information in report file. If a list of node names is provided, 
        EPANET only provides report information for those nodes.

    links : None, "ALL", or list
        Output link information in report file. If a list of link names is provided, 
        EPANET only provides report information for those links.

    pagesize : str
        Page size for EPANET report output
    

    """
    def __init__(self,
                pagesize: list=None,
                report_filename: str=None,
                status: str='NO',
                summary: str='YES',
                energy: str='NO',
                nodes: bool=False,
                links: bool=False,
                report_params: dict=None,
                param_opts: dict=None):
        self.pagesize = pagesize
        self.report_filename = report_filename
        self.status = status
        self.summary = summary
        self.energy = energy
        self.nodes = nodes
        self.links = links
        self.report_params = report_params if report_params is not None else _new_report_params()
        self.param_opts = param_opts if param_opts is not None else _new_param_opts()


class UserOptions(_OptionsBase):
    """
    Options defined by the user.
    
    Provides an empty class that accepts getattribute and setattribute methods to
    create user-defined options. For example, if using WNTR for uncertainty 
    quantification, certain options could be added here that would never be 
    used directly by WNTR, but which would be saved on pickling and could be
    used by the user-built analysis scripts.
    
    """
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            self.__dict__[k] = v


class Options(_OptionsBase):
    """
    Water network model options class.
    
    These options mimic options in EPANET.
    The `user` attribute is a generic python class object that allows for 
    dynamically created attributes that are user specific.

    Parameters
    ----------
    time : TimeOptions
        Contains all timing options for the scenarios

    hydraulic : HydraulicOptions
        Contains hydraulic solver parameters

    reaction : ReactionOptions
        Contains chemical reaction parameters

    quality : QualityOptions
        Contains water quality simulation options and source definitions

    energy : EnergyOptions
        Contains parameters for energy calculations

    report : ReportOptions
        Contains options for how for format and save report

    graphics : GraphicsOptions
        Contains EPANET graphics and background options and also the filename
        for external node coordinates, if used

    user : dict
        An empty dictionary that allows for user specified options
    
    
    """
    def __init__(self,
                 time: TimeOptions = None,
                 hydraulic: HydraulicOptions = None,
                 report: ReportOptions = None,
                 quality: QualityOptions = None,
                 reaction: ReactionOptions = None,
                 energy: EnergyOptions = None,
                 graphics: GraphicsOptions = None,
                 user: UserOptions = None):
        self.time = TimeOptions.factory(time)
        self.hydraulic = HydraulicOptions.factory(hydraulic)
        self.report = ReportOptions.factory(report)
        self.quality = QualityOptions.factory(quality)
        self.reaction = ReactionOptions.factory(reaction)
        self.energy = EnergyOptions.factory(energy)
        self.graphics = GraphicsOptions.factory(graphics)
        self.user = UserOptions.factory(user)

    def __setattr__(self, name, value):
        if name == 'time':
            if not isinstance(value, (TimeOptions, dict, tuple, list)):
                raise ValueError('time must be a TimeOptions or convertable object')
            value = TimeOptions.factory(value)
        elif name == 'hydraulic':
            if not isinstance(value, (HydraulicOptions, dict, tuple, list)):
                raise ValueError('hydraulic must be a HydraulicOptions or convertable object')
            value = HydraulicOptions.factory(value)
        elif name == 'report':
            if not isinstance(value, (ReportOptions, dict, tuple, list)):
                raise ValueError('report must be a ReportOptions or convertable object')
            value = ReportOptions.factory(value)
        elif name == 'quality':
            if not isinstance(value, (QualityOptions, dict, tuple, list)):
                raise ValueError('quality must be a QualityOptions or convertable object')
            value = QualityOptions.factory(value)
        elif name == 'reaction':
            if not isinstance(value, (ReactionOptions, dict, tuple, list)):
                raise ValueError('reaction must be a ReactionOptions or convertable object')
            value = ReactionOptions.factory(value)
        elif name == 'energy':
            if not isinstance(value, (EnergyOptions, dict, tuple, list)):
                raise ValueError('energy must be a EnergyOptions or convertable object')
            value = EnergyOptions.factory(value)
        elif name == 'graphics':
            if not isinstance(value, (GraphicsOptions, dict, tuple, list)):
                raise ValueError('graphics must be a GraphicsOptions or convertable object')
            value = GraphicsOptions.factory(value)
        elif name == 'user':
            value = UserOptions.factory(value)
        else:
            raise ValueError('%s is not a valid member of WaterNetworkModel')
        self.__dict__[name] = value


    def todict(self):
        """Dictionary representation of the options"""
        return dict(self)



"""
The wntr.scenario package contains methods to define disaster scenarios and 
fragility/survival curves.
"""
from wntr.scenario.earthquake import Earthquake
from wntr.scenario.fragility_curve import FragilityCurve


from matplotlib.colors import LinearSegmentedColormap
import logging

logger = logging.getLogger(__name__)

def custom_colormap(numcolors=11, colors=['blue','white','red'], name='custom'):
    """ 
    Create a custom colormap.  Default settings creates a colormap named 'custom'
    with 11 bins which transitions from blue to white to red.
    
    Parameters
    -----------
    numcolors : int (optional)
        Number of bins in the colormap.
        
    colors : list of colors (optional)
        Colors can be specified in any way understandable by 
        matplotlib.colors.ColorConverter.to_rgb().  
    
    name : str (optional)
        Name of the colormap
        
    Returns
    --------
    cmap : matplotlib.colors.LinearSegmentedColormap object
    """
    
    cmap = LinearSegmentedColormap.from_list(name=name, 
                                             colors = colors,
                                             N=numcolors)
    return cmap



"""
The wntr.graphics.curve module includes methods plot fragility curves and 
pump curves.
"""
import numpy as np
import matplotlib.pyplot as plt
import logging

logger = logging.getLogger(__name__)

def plot_fragility_curve(FC, fill=True, key='Default',
                         title='Fragility curve', 
                         xmin=0, xmax=1, npoints=100, 
                         xlabel='x', 
                         ylabel='Probability of exceeding a damage state',
                         ax=None):
    """
    Plot fragility curve.
    
    Parameters
    -----------
    FC : wntr.scenario.FragilityCurve object
        Fragility curve
    fill : bool, optional
        If true, fill area under the curve
    key : string, optional
        Fragility curve state distribution key
    title : string, optional
        Plot title
    xmin : float, optional
        X axis minimum 
    xmax : float, optional
        X axis maximum 
    npoints : int, optional
        Number of points 
    xlabel : string, optional
        X axis label
    ylabel : string, optional
        Y axis label
    ax : matplotlib axes object, optional
        Axes for plotting (None indicates that a new figure with a single 
        axes will be used)
        
    Returns
    ---------
    ax : matplotlib axes object
    """
    if ax is None: # create a new figure
        plt.figure(figsize=[8,4])
        ax = plt.gca()
        
    ax.set_title(title)
    
    x = np.linspace(xmin,xmax,npoints)
    for name, state in FC.states():
        try:
            dist=state.distribution[key]
            if fill:
                plt.fill_between(x, dist.cdf(x), label=name)
            else:
                plt.plot(x, dist.cdf(x), label=name)
        except:
            pass
        
    ax.set_xlim((xmin,xmax))
    ax.set_ylim((0,1))
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.legend()
    
    return ax

def plot_pump_curve(pump, title='Pump curve', 
                    xmin=0, xmax=None, ymin=0, ymax=None, 
                    xlabel='Flow (m$^3$/s)',
                    ylabel='Head (m)', 
                    ax=None):
    """
    Plot points in the pump curve along with the pump curve polynomial using 
    head curve coefficients (H = A - B*Q**C)
    
    Parameters
    -----------
    pump : wntr.network.elements.Pump object
        Pump
    title : string, optional
        Plot title
    xmin : float, optional
        X axis minimum
    xmax : float, optional
        X axis maximum
    ymin : float, optional
        Y axis minimum
    ymax : float, optional
        Y axis maximum
    xlabel : string, optional
        X axis label
    ylabel : string, optional
        Y axis label 
    ax : matplotlib axes object, optional
        Axes for plotting (None indicates that a new figure with a single 
        axes will be used)
        
    Returns
    ---------
    ax : matplotlib axes object  
    """
    try:
        curve = pump.get_pump_curve()
    except:
        print("Pump "+pump.name+" has no curve")
        return
    
    if ax is None: # create a new figure
        plt.figure(figsize=[8,4])
        ax = plt.gca()
        
    ax.set_title(title)
    
    cdata = np.array(curve.points)
    Q = cdata[:,0]
    H = cdata[:,1]

    try:
        coeff = pump.get_head_curve_coefficients()
        A = coeff[0]
        B = coeff[1]
        C = coeff[2]
        
        Q_max = (A/B)**(1/C)
        q = np.linspace(0, Q_max, 50)
        h = A - B*q**C
        
        ax.plot(q, h, '--', linewidth=1)
    except:
        pass
    
    ax.plot(Q, H, 'o', label=curve.name)    
    
    ax.set_xlim((xmin,xmax))
    ax.set_ylim((ymin,ymax))
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.legend()
    
    return ax

def plot_tank_volume_curve(tank, title='Tank volume curve', 
                    ax=None):
    """
    Plots a tank volume curve and the corresponding axi-symmetric tank profile shape
    
    Parameters
    -----------
    tank : wntr.network.elements.Tank object
        Tank 
    title : string, optional
        Plot title
    ax : matplotlib axes object, optional
        Axes for plotting (None indicates that a new figure with two subplots 
        will be used)
        
    Returns
    ---------
    ax : matplotlib axes object
        ax[0] is the left hand plot, ax[1] is the right hand plot
    """
    curve = tank.vol_curve
    
    if curve is None:
        print("Tank "+tank.name+" has no volume curve")
        return

    if ax is None: # create a new figure
        fig,ax = plt.subplots(1,2,figsize=[12,4])

    cdata = np.array(curve.points)
    L = cdata[:,0]
    V = cdata[:,1]
    
    V_min_level = tank.get_volume(tank.min_level)
    V_max_level = tank.get_volume(tank.max_level)

    ax[0].plot(L, V, '-o', linewidth=1,label="volume curve")
    ax[0].plot([tank.min_level,tank.min_level], [V_min_level,V_max_level],'-.',
        label="min level",color='r')
    ax[0].plot([tank.max_level,tank.max_level], [V_min_level,V_max_level],'-.',
        label="max level",color='r')
    ax[0].grid("on")
    ax[0].set_xlabel("Tank level (m)")
    ax[0].set_ylabel("Tank volume (m$^3$)")
    ax[0].set_title(title)
    ax[0].legend()
    
    # calculate the tank profile assuming an axi-symmetric tank
    d = []
    l = []
    #d.append(0.0)
    #d.append(tank.diameter)
    #d.append(tank.diameter)
    #l.append(0.0)
    #l.append(0.0)
    #l.append(tank.min_level)
    lev0 = L[0]
    vol0 = V[0]
    for vol,lev in zip(V[1:],L[1:]):
        dn = np.sqrt(4.0*(vol-vol0)/(np.pi * (lev-lev0)))
        l.append(lev0)
        l.append(lev)
        d.append(dn)
        d.append(dn)
        lev0 = lev
        vol0 = vol
    #l.append(l[-1])
    #d.append(0.0)
    
    ax[1].plot(np.array(d)/2,l,label="tank profile")
    max_d = max([tank.diameter,max(d)])
    ax[1].plot([0.0,max_d/2.0],[tank.min_level,tank.min_level],'-.',
        label="min level",color='r')
    ax[1].plot([0.0,max_d/2.0],[tank.max_level,tank.max_level],'-.',
        label="max level",color='r')
    #ax[1].plot([0.0,max_d/2.0],[0.0,0.0],'-.',label='elevation={0:5.2f}m'.format(tank.elevation),color='k')
    ax[1].grid("on")
    ax[1].set_xlabel("Equivalent axisymmetric tank radius (m)")
    ax[1].set_ylabel("Tank level (m)")
    ax[1].set_title("Geometric tank profile")
    
    xlim = list(ax[1].get_xlim())
    ylim = list(ax[1].get_ylim())
    if np.diff(xlim) > np.diff(ylim):
        ylim[1] = ylim[0] + (xlim[1]-xlim[0])
    else:
        xlim[1] = xlim[0] + (ylim[1] - ylim[0])
    ax[1].set_xlim(xlim)
    ax[1].set_ylim(ylim)
    ax[1].legend()
    
    return ax



"""
The wntr.metrics.topographic module contains topographic metrics that are not
available directly with NetworkX.  Functions in this module operate on a 
NetworkX MultiDiGraph, which can be created by calling ``G = wn.get_graph()``

.. rubric:: Contents

.. autosummary::

    terminal_nodes
    bridges
    central_point_dominance
    spectral_gap
    algebraic_connectivity
    critical_ratio_defrag
    valve_segments

"""
import networkx as nx
import numpy as np
import pandas as pd
import logging

logger = logging.getLogger(__name__)

def terminal_nodes(G):
    """
    Nodes with degree 1

    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph

    Returns
    -------
    List of terminal nodes
    
    """
    node_degree = dict(G.degree())
    terminal_nodes = [k for k,v in node_degree.items() if v == 1]

    return terminal_nodes

def bridges(G):
    """
    Bridge links

    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph

    Returns
    -------
    List of links that are bridges
    
    """
    uG = G.to_undirected() # uses an undirected graph
    bridge_links = []
    bridges = nx.bridges(nx.Graph(uG)) # not implemented for multigraph
    for br in bridges:
        for key in uG[br[0]][br[1]].keys():
            bridge_links.append(key)
        
    return bridge_links

def central_point_dominance(G):
    """
    Central point dominance
    
    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph
        
    Returns
    -------
    Central point dominance (float)
    
    """
    uG = G.to_undirected() # uses an undirected graph
    bet_cen = nx.betweenness_centrality(nx.Graph(uG)) # not implemented for multigraph 
    bet_cen = list(bet_cen.values())
    cpd = sum(max(bet_cen) - np.array(bet_cen))/(len(bet_cen)-1)

    return cpd

def spectral_gap(G):
    """
    Spectral gap
    
    Difference in the first and second eigenvalue of the adjacency matrix
    
    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph
        
    Returns
    -------
    Spectral gap (float)
    
    """
    uG = G.to_undirected() # uses an undirected graph
    eig = nx.adjacency_spectrum(uG)
    spectral_gap = abs(eig[0] - eig[1])

    return spectral_gap.real

def algebraic_connectivity(G):
    """
    Algebraic connectivity
    
    Second smallest eigenvalue of the normalized Laplacian matrix of a network

    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph
        
    Returns
    -------
    Algebraic connectivity (float)
    
    """
    uG = G.to_undirected() # uses an undirected graph
    eig = nx.laplacian_spectrum(uG)
    eig = np.sort(eig)
    alg_con = eig[1]

    return alg_con

def critical_ratio_defrag(G):
    """
    Critical ratio of defragmentation

    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph
        
    Returns
    -------
    Critical ratio of defragmentation (float)
    
    """
    node_degree = dict(G.degree())
    tmp = np.mean(pow(np.array(list(node_degree.values())),2))
    fc = 1-(1/((tmp/np.mean(list(node_degree.values())))-1))

    return fc


def _links_in_simple_paths(G, sources, sinks):
    """
    Count all links in a simple path between sources and sinks

    Parameters
    -----------
    G: networkx MultiDiGraph
        Graph
    sources: list
        List of source nodes
    sinks: list
        List of sink nodes

    Returns
    -------
    Dictionary with the number of times each link is involved in a path
    
    """
    link_names = [name for (node1, node2, name) in list(G.edges(keys=True))]
    link_count = pd.Series(data = 0, index=link_names)

    for sink in sinks:
        for source in sources:
            if nx.has_path(G, source, sink):
                paths = nx.all_simple_paths(G,source,target=sink)
                for path in paths:
                    for i in range(len(path)-1):
                        links = list(G[path[i]][path[i+1]].keys())
                        for link in links:
                            link_count[link] = link_count[link]+1

    return link_count

def valve_segments(G, valve_layer):
    """
    Valve segmentation

    Parameters
    -----------
    G: networkx MultiDiGraph
        Graph
    valve_layer: pandas DataFrame
        Valve layer, defined by node and link pairs (for example, valve 0 is 
        on link A and protects node B). The valve_layer DataFrame is indexed by
        valve number, with columns named 'node' and 'link'.

    Returns
    -------
    node_segments: pandas Series
       Segment number for each node, indexed by node name
    link_segments: pandas Series
        Segment number for each link, indexed by link name
    segment_size: pandas DataFrame
        Number of nodes and links in each segment. The DataFrame is indexed by 
        segment number, with columns named 'node' and 'link'.
    """
    # Convert the graph to an undirected graph
    uG = G.to_undirected()
    
    # Node and link names
    nodes = list(uG.nodes()) # list of node names
    links = list(uG.edges(keys=True)) # list of tuples with start node, end node, link name
    
    # Append N_ and L_ to node and link names, used in matrices
    matrix_node_names = ['N_'+n for n in nodes]
    matrix_link_names = ['L_'+k for u,v,k in links]

    # Pipe-node connectivity matrix
    A = nx.incidence_matrix(uG).todense().T
    AC = pd.DataFrame(A, columns=matrix_node_names, index=matrix_link_names, dtype=int)
    
    # Valve-node connectivity matrix
    VC = pd.DataFrame(0, columns=matrix_node_names, index=matrix_link_names)
    for i, row in valve_layer.iterrows():
        VC.at['L_'+row['link'], 'N_'+row['node']] = 1
    
    # Valve deficient matrix (anti-valve matrix)
    VD = AC - VC

    # Direct connectivity matrix
    NI = pd.DataFrame(np.identity(len(matrix_node_names)),
                      index = matrix_node_names, columns = matrix_node_names)
    LI = pd.DataFrame(np.identity(len(matrix_link_names)),
                      index = matrix_link_names, columns = matrix_link_names)
    DC_left = pd.concat([NI, VD], sort=False)
    DC_right = pd.concat([VD.T, LI], sort=False)
    DC = pd.concat([DC_left, DC_right], axis=1, sort=False)
    DC = DC.astype(int)

    # initialization for looping routine
    seg_index = 0
    
    # pre-processing to find isolated elements before looping
    seg_label = {}

    for start_node, end_node, link_name in links:
        link_valves = valve_layer[valve_layer['link']==link_name]
        if set(link_valves['node']) >= set([start_node, end_node]):
            seg_index += 1
            seg_label['L_'+link_name] = seg_index
            
    for node_name in nodes:
        node_valves = valve_layer[valve_layer['node']==node_name]
        node_links = [k for u,v,k in uG.edges(node_name, keys=True)]
        if set(node_valves['link']) >= set(node_links):
            seg_index += 1
            seg_label['N_'+node_name] = seg_index
    
    # drop previously found isolated elements before looping
    DC = DC.drop(seg_label.keys())
    DC = DC.drop(seg_label.keys(), axis=1)   
    
    DC_np = DC.to_numpy() # requires Pandas v.0.24.0

    # vector of length nodes+links where the ith entry is the segment number of node/link i
    seg_label_DC = np.zeros(shape=(len(DC.index)), dtype=int)

    # Loop over all nodes and links to grow segments
    for i in range(len(seg_label_DC)):
        
        # Only assign a seg_label if node/link doesn't already have one
        if seg_label_DC[i] == 0:
            
            # Advance segment label and assign to node/link, mark as assigned
            seg_index += 1
            seg_label_DC[i] = seg_index
           
            # Initialize segment size
            seg_size = (seg_label_DC == seg_index).sum()
             
            flag = True
            
            #print(i)
    
            # Nodes and links that are part of the segment
            seg = np.where(seg_label_DC == seg_index)[0]
        
            # Connectivity of the segment
            seg_DC = np.zeros(shape=(DC_np.shape[0]), dtype=int)
            seg_DC[seg] = 1
    
            while flag:          
               
                # Potential connectivity of the segment      
                p_seg_DC = DC_np + seg_DC[:,None] # this is slow
                
                # Nodes and links that are connected to the segment
                temp = np.max(p_seg_DC,axis=0) # this is somewhat slow
                connected_to_seg = np.where(temp > 1)[0]   
                seg_DC[connected_to_seg] = 1
      
                # Label nodes/links connected to the segment
                seg_label_DC[connected_to_seg] = seg_index
                
                # Find new segment size
                new_seg_size = (seg_label_DC == seg_index).sum()
                    
                # Check for progress
                if seg_size == new_seg_size:
                    flag = False
                else:
                    seg_size = new_seg_size
          
                # Update seg_DC and DC_np
                seg_DC = DC_np[:,i] + np.sum(
                                        p_seg_DC[:,connected_to_seg],axis=1
                                        ) # this is slow
                seg_DC = np.clip(seg_DC,0,1)          
                DC_np[:,connected_to_seg] = np.repeat(
                                            seg_DC,len(connected_to_seg)).reshape(
                                            len(seg_DC),len(connected_to_seg))
                                            # this is somewhat slow
    
    #        print(i, seg_size)
    
    # combine pre-processed and looped results
    seg_labels = list(seg_label.values()) + list(seg_label_DC)
    seg_labels_index = list(seg_label.keys()) + list(DC.index)
    seg_label = pd.Series(seg_labels, index=seg_labels_index, dtype=int)

    # Separate node and link segments
    # remove leading N_ and L_ from node and link names
    node_segments = seg_label[matrix_node_names]
    link_segments = seg_label[matrix_link_names]
    node_segments.index = node_segments.index.str[2::]
    link_segments.index = link_segments.index.str[2::]  
    
    # Extract segment sizes, for nodes and links
    seg_link_sizes = link_segments.value_counts().rename('link')
    seg_node_sizes = node_segments.value_counts().rename('node')
    seg_sizes = pd.concat([seg_link_sizes, seg_node_sizes], axis=1).fillna(0)
    seg_sizes = seg_sizes.astype(int)
    
    return node_segments, link_segments, seg_sizes




"""
The wntr.morph.node module contains functions to modify node coordinates.
"""
import logging
import copy
import numpy as np
from scipy.spatial.distance import pdist
try:
    import utm
except:
    utm = None

logger = logging.getLogger(__name__)


def scale_node_coordinates(wn, scale, return_copy=True):
    """
    Scales node coordinates, using 1:scale
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        Water network model
    scale: float
        Coordinate scale multiplier, in meters
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    --------
    wntr WaterNetworkModel
        Water network model with updated node coordinates
    """
    if return_copy: # Get a copy of the WaterNetworkModel
        wn2 = copy.deepcopy(wn)
    else:
        wn2 = wn
    
    for name, node in wn2.nodes():
        pos = node.coordinates
        node.coordinates = (pos[0]*scale, pos[1]*scale)

    return wn2


def translate_node_coordinates(wn, offset_x, offset_y, return_copy=True):
    """
    Translate node coordinates
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        Water network model
    offset_x: tuple
        Translation in the x direction, in meters
    offset_y: float
        Translation in the y direction, in meters
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    --------
    wntr WaterNetworkModel
        Water network model with updated node coordinates
    """
    if return_copy: # Get a copy of the WaterNetworkModel
        wn2 = copy.deepcopy(wn)
    else:
        wn2 = wn
    
    for name, node in wn2.nodes():
        pos = node.coordinates
        node.coordinates = (pos[0]+offset_x, pos[1]+offset_y)
    
    return wn2
        

def rotate_node_coordinates(wn, theta, return_copy=True):
    """
    Rotate node coordinates counter-clockwise by theta degrees
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        Water network model
    theta: float
        Node rotation, in degrees
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    --------
    wntr WaterNetworkModel
        Water network model with updated node coordinates
    """
    if return_copy: # Get a copy of the WaterNetworkModel
        wn2 = copy.deepcopy(wn)
    else:
        wn2 = wn
    
    theta = np.radians(theta)
    R = np.array([[np.cos(theta),-np.sin(theta)], 
                  [np.sin(theta), np.cos(theta)]])
    for name, node in wn2.nodes():
        pos = node.coordinates
        node.coordinates = tuple(np.dot(R,pos))
    
    return wn2


def convert_node_coordinates_UTM_to_longlat(wn, zone_number, zone_letter, return_copy=True):
    """
    Convert node coordinates from UTM coordinates to longitude, latitude coordinates
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        Water network model
    zone_number: int
       Zone number
    zone_letter: string
        Zone letter
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    --------
    wntr WaterNetworkModel
        Water network model with updated node coordinates (longitude, latitude)
    """
    if utm is None:
        raise ImportError('utm package is required')
    
    if return_copy: # Get a copy of the WaterNetworkModel
        wn2 = copy.deepcopy(wn)
    else:
        wn2 = wn
    
    for name, node in wn2.nodes():
        pos = node.coordinates
        lat, long = utm.to_latlon(pos[0], pos[1], zone_number, zone_letter)
        node.coordinates = (long, lat)

    return wn2


def convert_node_coordinates_longlat_to_UTM(wn, return_copy=True):
    """
    Convert node longitude, latitude coordinates to UTM coordinates
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        Water network model
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    --------
    wntr WaterNetworkModel
        Water network model with updated node coordinates (easting, northing)
    """
    if utm is None:
        raise ImportError('utm package is required')
    
    if return_copy: # Get a copy of the WaterNetworkModel
        wn2 = copy.deepcopy(wn)
    else:
        wn2 = wn
    
    for name, node in wn2.nodes():
        pos = node.coordinates
        longitude = pos[0]
        latitude = pos[1]
        utm_coords = utm.from_latlon(latitude, longitude)
        easting = utm_coords[0]
        northing = utm_coords[1]
        node.coordinates = (easting, northing)

    return wn2


def convert_node_coordinates_to_UTM(wn, utm_map, return_copy=True):
    """
    Convert node coordinates to UTM coordinates
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        Water network model
    utm_map: dictionary
        Dictionary containing two node names and their x, y coordinates in 
        UTM easting, northing in the format
        {'node name 1': (easting, northing), 'node name 2': (easting, northing)}
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    --------
    wntr WaterNetworkModel
        Water network model with updated node coordinates (easting, northing)
    """
    
    wn2 = _convert_with_map(wn, utm_map, 'UTM', return_copy)
    
    return wn2


def convert_node_coordinates_to_longlat(wn, longlat_map, return_copy=True):
    """
    Convert node coordinates to longitude, latitude coordinates
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        Water network model
    longlat_map: dictionary
        Dictionary containing two node names and their x, y coordinates in
        longitude, latitude in the format
        {'node name 1': (longitude, latitude), 'node name 2': (longitude, latitude)}
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    --------
    wntr WaterNetworkModel
        Water network model with updated node coordinates (longitude, latitude)
    """
    
    wn2 = _convert_with_map(wn, longlat_map, 'LONGLAT', return_copy)
    
    return wn2

def _convert_with_map(wn, node_map, flag, return_copy):
    
    if utm is None:
        raise ImportError('utm package is required')
    
    if not len(node_map.keys()) == 2:
        print('map must have exactly 2 entries')
        return
    
    if return_copy: # Get a copy of the WaterNetworkModel
        wn2 = copy.deepcopy(wn)
    else:
        wn2 = wn
    
    node_names = list(node_map.keys())
    
    A = []
    B = []
    for node_name, coords in node_map.items():
        A.append(np.array(wn2.get_node(node_name).coordinates))
        if flag == 'LONGLAT':
            longitude = coords[0]
            latitude = coords[1]
            utm_coords = utm.from_latlon(latitude, longitude)
            zone_number = utm_coords[2]
            zone_letter = utm_coords[3] 
            B.append(np.array(utm_coords[0:2])) # B is in UTM coords
        elif flag == 'UTM':
            easting = coords[0]
            northing = coords[1]
            B.append(np.array(easting, northing)) # B is in UTM coords
    
    # Rotate, if needed
    vect1 = A[1] - A[0]
    vect2 = B[1] - B[0]
    vect1_unit = vect1/np.linalg.norm(vect1)
    vect2_unit = vect2/np.linalg.norm(vect2)
    dotproduct = np.dot(vect1_unit, vect2_unit)
    if dotproduct < 1:
        sign = np.sign(np.cross(vect1_unit, vect2_unit))
        angle = np.arccos(dotproduct)*180/np.pi
        wn2 = rotate_node_coordinates(wn2, sign*angle)
        A[0] = np.array(wn2.get_node(node_names[0]).coordinates)
        A[1] = np.array(wn2.get_node(node_names[1]).coordinates)
        
    # Compute center points
    cpA = np.mean(A, axis=0)
    cpB = np.mean(B, axis=0)
    
    # Compute distance to each center point
    distA = np.mean([pdist([A[0], cpA])[0], pdist([A[1], cpA])[0]])     
    distB = np.mean([pdist([B[0], cpB])[0], pdist([B[1], cpB])[0]])
    
    # Compute ratio
    ratio = distB/distA
    
    for name, node in wn2.nodes():
        pos = node.coordinates
        (easting, northing) = (np.array(pos) - cpA)*ratio + cpB
        if flag == 'LONGLAT':
            lat, long = utm.to_latlon(easting, northing, zone_number, zone_letter)
            node.coordinates = (long, lat)
        elif flag == 'UTM':
            node.coordinates = (easting, northing)
    
    return wn2


"""
The wntr.network.layer module includes methods to generate network layers
(information that is not stored in the water network model or the graph).

.. rubric:: Contents

.. autosummary::

    generate_valve_layer
"""
import numpy as np
import pandas as pd

def generate_valve_layer(wn, placement_type='strategic', n=1, seed=None):
    """
    Generate valve layer data, which can be used in valve segmentation analysis.

    Parameters
    -----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object
        
    placement_type : string
        Options include 'strategic' and 'random'.  
        
        - If 'strategic', n is the number of pipes from each node that do not 
          contain a valve. In this case, n is generally 0, 1 or 2 
          (i.e. N, N-1, N-2 valve placement).
        - If 'random', then n randomly placed valves are used to define the 
          valve layer.
        
    n : int
        
        - If 'strategic', n is the number of pipes from each node that do not 
          contain a valve.
        - If 'random', n is the number of number of randomly placed valves.
        
    seed : int or float
        Random seed
       
    Returns
    ---------
    valve_layer : pandas DataFrame
        Valve layer, defined by node and link pairs (for example, valve 0 is 
        on link A and protects node B). The valve_layer DataFrame is indexed by
        valve number, with columns named 'node' and 'link'.
    """
    
    if seed is not None:
        np.random.seed(seed)
    
    valve_layer = []
    if placement_type=='random':
        for pipe_name in np.random.choice(wn.pipe_name_list, n):
            pipe = wn.get_link(pipe_name)
            valve_layer.append([pipe_name, pipe.start_node_name])
            
    elif placement_type == 'strategic':
        for node_name, node in wn.nodes():
            links = wn.get_links_for_node(node_name)
            for l in np.random.choice(links, max(len(links)-n,0), replace=False):
                valve_layer.append([l, node_name])
            
    valve_layer = pd.DataFrame(valve_layer, columns=['link', 'node'])  
    
    return valve_layer